"""Interface to Q-Chem, the ab initio quantum chemistry package"""
import numpy as np
import xtal
from ezff.utils import convert_units as convert



def read_structure(outfilename):
    """
    Read-in a multiple partially-converged structures from a PES scan (including bond-scans, angle-scans and dihedral-scans)

    :param outfilename: Single filename for ``stdout`` from the QChem PES scan job or a list of filenames for ``stdout`` files from partial QChem PES scan jobs
    :type outfilename: str

    :returns: ``xtal`` trajectory object with structures and converged energies along the PES scan as individual snapshots
    """
    structure = xtal.AtTraj()
    structure.box = np.zeros((3,3))
    energies = []

    if isinstance(outfilename, list):
        listfilenames = outfilename
    else:
        listfilenames = [outfilename]

    for single_outfilename in list(listfilenames):
        # Read structure
        outfile = open(single_outfilename,'r')
        while True:
            line = outfile.readline()
            if not line: # break at EOF
                break
            if 'OPTIMIZATION CONVERGED' in line:
                snapshot = structure.create_snapshot(xtal.Snapshot)
                dummyline = outfile.readline()
                dummyline = outfile.readline()
                dummyline = outfile.readline()
                dummyline = outfile.readline()
                # Atomic coordinates start here
                while True:
                    coords = outfile.readline()
                    if coords.strip()=='':
                        break
                    atom = snapshot.create_atom(xtal.Atom)
                    atom.element, atom.cart = coords.strip().split()[1], np.array(list(map(float,coords.strip().split()[2:5])))
                xpos_this_snapshot = [atom.cart[0] for atom in snapshot.atomlist]
                ypos_this_snapshot = [atom.cart[1] for atom in snapshot.atomlist]
                zpos_this_snapshot = [atom.cart[2] for atom in snapshot.atomlist]
                offset_vector = 0.0 - np.array([np.amin(xpos_this_snapshot), np.amin(ypos_this_snapshot), np.amin(zpos_this_snapshot)])
                snapshot.move(offset_vector)
        outfile.close()

    # Since the QChem structure is often non-periodic, we set up a box that is 50 Angstrom larger than the maximum extent of atoms
    xpos = [atom.cart[0] for atom in snapshot.atomlist for snapshot in structure.snaplist]
    ypos = [atom.cart[1] for atom in snapshot.atomlist for snapshot in structure.snaplist]
    zpos = [atom.cart[2] for atom in snapshot.atomlist for snapshot in structure.snaplist]

    structure.box[0][0] = np.amax(xpos) - np.amin(xpos) + 50.0
    structure.box[1][1] = np.amax(ypos) - np.amin(ypos) + 50.0
    structure.box[2][2] = np.amax(zpos) - np.amin(zpos) + 50.0

    return structure



def read_energy(outfilename):
    """
    Read-in a multiple partially-converged structures from a PES scan (including bond-scans, angle-scans and dihedral-scans)

    :param outfilename: Single filename for ``stdout`` from the QChem PES scan job or a list of filenames for ``stdout`` files from partial QChem PES scan jobs
    :type outfilename: str

    :returns: ``xtal`` trajectory object with structures and converged energies along the PES scan as individual snapshots
    """
    structure = xtal.AtTraj()
    energies = []

    if isinstance(outfilename, list):
        listfilenames = outfilename
    else:
        listfilenames = [outfilename]

    for single_outfilename in list(listfilenames):
        # Read energies
        outfile = open(single_outfilename,'r')
        for line in outfile:
            if 'Final energy is' in line:
                energy_in_Hartrees = float(line.strip().split()[-1])
                energies.append(energy_in_Hartrees * convert.energy['Ha']['eV'])
        outfile.close()
    return np.array(energies)



def read_atomic_charges(outfilename):
    """
    Read-in a multiple partially-converged structures from a PES scan (including bond-scans, angle-scans and dihedral-scans)

    :param outfilename: Single filename for ``stdout`` from the QChem PES scan job or a list of filenames for ``stdout`` files from partial QChem PES scan jobs
    :type outfilename: str

    :returns: ``xtal`` trajectory object with structures and converged energies along the PES scan as individual snapshots
    """
    structure = xtal.AtTraj()
    structure.box = np.zeros((3,3))
    energies = []

    if isinstance(outfilename, list):
        listfilenames = outfilename
    else:
        listfilenames = [outfilename]

    for single_outfilename in list(listfilenames):
        # Read structure
        outfile = open(single_outfilename,'r')
        while True:
            line = outfile.readline()
            if not line: # break at EOF
                break

            if 'Ground-State Mulliken Net Atomic Charges' in line:
                charge_array = []
                dummyline = outfile.readline()
                dummyline = outfile.readline()
                dummyline = outfile.readline()
                while True:
                    charges = outfile.readline().strip().split()
                    if charges[0][0] == '-':
                        break
                    charge_array.append(float(charges[2]))

            if 'OPTIMIZATION CONVERGED' in line:
                snapshot = structure.create_snapshot(xtal.Snapshot)
                dummyline = outfile.readline()
                dummyline = outfile.readline()
                dummyline = outfile.readline()
                dummyline = outfile.readline()
                # Atomic coordinates start here
                atomID = 0
                while True:
                    coords = outfile.readline()
                    if coords.strip()=='':
                        break
                    atom = snapshot.create_atom(xtal.Atom)
                    atom.element, atom.cart = coords.strip().split()[1], np.array(list(map(float,coords.strip().split()[2:5])))
                    atom.charge = charge_array[atomID]
                    atomID += 1

        outfile.close()
    return structure



import ezff
import xtal
from ezff.interfaces import qchem
import pytest
import numpy as np

def test_qchem_read_energy_singlepoint():
    energy = qchem.read_energy('qchem_output_files/optCHOSx.out')
    is_correct_type = isinstance(energy, np.ndarray)
    is_correct_size = (len(energy) == 1)
    is_correct_value = (np.linalg.norm(energy) == pytest.approx(14044.3073))
    assert (is_correct_type and is_correct_size and is_correct_value)


def test_qchem_read_structure_singlepoint():
    structure = qchem.read_structure('qchem_output_files/optCHOSx.out')
    has_correct_snapsize = (len(structure.snaplist) == 1)
    has_correct_atomsize = (len(structure.snaplist[0].atomlist) == 10)
    has_correct_data_1 = (np.linalg.norm(structure.snaplist[0].atomlist[0].cart - structure.snaplist[0].atomlist[1].cart) == pytest.approx(1.6389119))
    assert (has_correct_snapsize and has_correct_atomsize and has_correct_data_1)



"""Interface to RXMD, the General Utility Lattice Program"""
import os
import xtal
import numpy as np
from ezff.utils import convert_units as convert

class job:
    """
    Class representing a RXMD calculation
    """

    def __init__(self, verbose=False, path='.'):
        """
        :param path: Path where the RXMD job must be run from
        :type path: str

        :param verbose: Print details about the RXMD job
        :type verbose: bool
        """
        if not os.path.isdir(path):
            if verbose:
                print('Path for current job is not valid . Creating a new directory...')
            os.makedirs(path)

        self.path = path
        
        self.outfile = self.path + '/log'
        print(self.outfile)
        self.command = 'rxmd'
        self.forcefield = ''
        self.structure = None

        self.verbose = verbose
        if verbose:
            print('Created a new RXMD job')

    def run(self, command = None, timeout = None):

        """
        Execute rxmd job with user-defined parameters

        :param command: path to RXMD executable
        :type command: str

        :param parallel: Flag for parallel execution
        :type parallel: bool

        :param processors: Number of processors for parallel execution of each RXMD job
        :type processors: int

        :param timeout: RXMD job is automatically killed after ``timeout`` seconds
        :type timeout: int
        """

        if command is None:
            command = self.command
        print(self.outfile)
        system_call_command = command + ' > ' + self.outfile + ' 2> ' + self.outfile + '.runerror'

        if timeout is not None:
            system_call_command = 'timeout ' + str(timeout) + ' ' + system_call_command

        # Create workspace DAT
        makeDir = 'mkdir -p ' + self.path + '/DAT'
        os.system(makeDir)

        # Copy rxmd.in file to path
        copyCommand1 = 'cp rxmd.in ' + self.path  
        copyCommand2 = 'cp rxff_nompi ' + self.path + '/DAT/.'
        

        self.write_script_file()

        os.system(copyCommand1)
        os.system(copyCommand2)        

        if self.verbose:
            print('cd '+ self.path + ' ; ' + system_call_command)

        os.system('cd '+ self.path + ' ; ' + system_call_command)

    def getAtomNames(self,ffieldString, isLG=False):

        atomNames = []
        counter = 0
        #with open(ffieldName, 'r') as ff:
        #ff.readline()
        print('<-----------------------------' + ffieldString[1] + '---------------------------------------->')
        numParams = int(ffieldString[1].split()[0])
        counter += 1
        for i in range(numParams):
            counter += 1

        counter += 1
        numAtomNames = int(ffieldString[counter].split()[0])
        counter += 3

        for i in range(numAtomNames):
            counter += 1
            atomNames.append(ffieldString.split()[0])
            counter += 3

            #if (isLG): ff.readline()
                #print('%d ----> %s' %(i+1, atomNames[-1]))

        print(atomNames)
        return atomNames

    def getBox(self,la,lb,lc,angle1,angle2,angle3):

        H  = np.zeros((3,3))
        Hi = np.zeros((3,3))

        lal = angle1 * (np.pi / 180.0)
        lbe = angle2 * (np.pi / 180.0)
        lga = angle3 * (np.pi / 180.0)

        hh1 = lc * (np.cos(lal) - np.cos(lbe)* np.cos(lga))/ np.sin(lga)
        hh2 = lc * pow(1.0 - pow(np.cos(lal),2) - pow(np.cos(lbe),2) - pow(np.cos(lga),2) + 2 * np.cos(lal) * np.cos(lbe) * np.cos(lga), 0.5)/np.sin(lga)

        H[0,0] = la
        H[1,0] = 0.0
        H[2,0] = 0.0
        H[0,1] = lb * np.cos(lga)
        H[1,1] = lb * np.sin(lga)
        H[2,1] = 0.0
        H[0,2] = lc * np.cos(lbe)
        H[1,2] = hh1
        H[2,2] = hh2

        print('---------------Hmatrix------------------')
        print('%12.6f   %12.6f    %12.6f' %(H[0,0], H[0,1], H[0,2]))
        print('%12.6f   %12.6f    %12.6f' %(H[1,0], H[1,1], H[1,2]))
        print('%12.6f   %12.6f    %12.6f' %(H[2,0], H[2,1], H[2,2]))

        Hi = np.linalg.inv(H)

        print('---------------Hinv--------------------')
        print('%12.6f   %12.6f    %12.6f' %(Hi[0,0], Hi[0,1], Hi[0,2]))
        print('%12.6f   %12.6f    %12.6f' %(Hi[1,0], Hi[1,1], Hi[1,2]))
        print('%12.6f   %12.6f    %12.6f' %(Hi[2,0], Hi[2,1], Hi[2,2]))

        return (H, Hi)

    def write_script_file(self, convert_reaxff=None):

        """
        Prepare input coordinate file to be read by rxmd code
        """

        assert len(self.structure.snaplist[0].atomlist) > 0, "No atoms in structure list"

      
        # To store input file data for rxmd
        #print(self.forcefield)
        #atomNames = self.getAtomNames(self.forcefield)
        atomNames = ['C', 'H', 'O', 'S', 'F', 'Cl', 'N']
        atomLen = len(atomNames)

        natoms = len(self.structure.snaplist[0].atomlist)
        l1, l2, l3, lalpha, lbeta, lgamma = None, None, None, None, None, None
        ctype0, pos, pos0,  itype0, itype1 = [], None, None, None, None

        rr = np.zeros((3,1))
        lbox = np.zeros((3,1))
        obox = np.zeros((3,1))
        H, Hi = np.zeros((3,3)), np.zeros((3,3))

        minX, minY, minZ = 9999, 9999, 9999

        ctype0 = []
        pos    = np.zeros((natoms,3), np.float64)
        pos0   = np.zeros((natoms,3), np.float64)
        itype0 = np.zeros((natoms, ), np.float64)
        itype1 = np.zeros((natoms, ), np.float64)
        
        i=0

        for i, atom in enumerate(self.structure.snaplist[0].atomlist):
            atype = atom.element.title()
            x, y, z = atom.cart[0], atom.cart[1], atom.cart[2]
 
            # Find the minimum coordinate in x, y and z dir and shift the coordinates later such that the origin of box is at 0,0,0
            if x < minX: minX = x
            if y < minY: minY = y
            if z < minZ: minZ = z
 
            pos[i] = np.array([x,y,z])
            #i=i+1

            for j in range(atomLen):
                if atype == atomNames[j]:
                    itype0[i] = j+1
                    break

            # Big floating number to store atom type, atomID, to be read by rxmd code
            itype1[i] = itype0[i] + pow(10,-13)*(i+1) + pow(10,-14)

        # Shift Origin
        pos = pos - np.array([minX, minY, minZ])        
        
        # Find box size

        
        l1, l2, l3 = np.max(pos[:,0]) + 10.0, np.max(pos[:,0]) + 10.0, np.max(pos[:,0]) + 10.0
        lalpha, lbeta, lgamma = 90, 90, 90 
        # Shift all the positions by 0.5
        pos = pos + 5.0

        # Normalize the coordinates by Hi matrix
        H, Hi = self.getBox(l1, l2, l3, lalpha, lbeta, lgamma)

        # Normalized structure is stored in pos0
        pos0 = np.dot(pos, Hi)

        # Wrap backtype0[i] coordinates contained in pos0 if they are out of box
    
        #rmin[0], rmin[1], rmin[2] = np.min(pos0[:,0]), np.min(pos0[:,1]), np.min(pos0[:,2])

        pos0 = pos0 - np.min(pos0,axis=0)   

        # Wrap back more if necessary
        pos0[:,0] = pos0[:,0]%1.0
        pos0[:,1] = pos0[:,1]%1.0
        pos0[:,2] = pos0[:,2]%1.0

        # Shift by a small amount to avoid coordinates at zero
        pos0 = pos0 + pow(10,-9)

        # Check minimum and maximum values of coordinates in pos0
        rmin = np.min(pos0,axis=1)
        rmax = np.max(pos0,axis=1)

        if self.verbose:
            print('rmin ------------> %12.6f  %12.6f  %12.6f' %(rmin[0], rmin[1], rmin[2]))
            print('rmax ------------> %12.6f  %12.6f  %12.6f' %(rmax[0], rmax[1], rmax[2]))

        # Write final coordinates to rxff_nompi to be read by rxmd, also write to geninit1.xyz for verification
        file1 = open('geninit1.xyz', 'w')    
        file2 = open('rxff_nompi', 'w')

        file1.write('%d\n' %(natoms))
        file2.write('1 1 1 1 %d 0\n' %(natoms))

        file1.write('%12.6f %12.6f %12.6f %12.6f %12.6f %12.6f\n' %(l1,l2,l3,lalpha,lbeta,lgamma))
        file2.write('%12.6f %12.6f %12.6f %12.6f %12.6f %12.6f\n' %(l1,l2,l3,lalpha,lbeta,lgamma))

        vv = np.zeros((natoms,3))
        rr1 = np.zeros((3,1))
        q = np.zeros((natoms,1))
        qfsp = 0.0
        qfsv = 0.0

        for n in range(natoms):
            rr1 = H.dot(pos0[n])
            dtype = itype1[n]
            gid = int((dtype - int(dtype))*pow(10,13))
            ity = int(dtype)-1
            file1.write('%s %12.6f %12.6f %12.6f %6d\n' %(atomNames[ity], rr1[0], rr1[1], rr1[2], gid))
            file2.write('%12.6f %12.6f %12.6f %12.6f %12.6f %12.6f %12.6f %16.14f %12.6f %12.6f\n'%(pos0[n][0],pos0[n][1],pos0[n][2],vv[n][0],vv[n][1],vv[n][2],q[n],dtype,qfsp,qfsv))
        
        file1.close()
        file2.close()

    def cleanup(self):
        """
        Clean-up after the completion of a RXMD job. Deletes input, output and forcefields files
        """
        files_to_be_removed = [self.outfile, self.outfile+'.runerror']
        for file in files_to_be_removed:

            if os.path.isfile(file):
                rm_command = 'rm -rf %s ' %file
                os.system(rm_command)
            elif os.path.isfile(self.path+'/'+file):
                rm_command = 'rm -rf %s ' %(self.path + '/' + file)
                os.system(rm_command)

        os.system('rm -rf ' + self.path + '/DAT')
        os.system('rm -rf ' + self.path + '/ffield')
        os.system('rm -rf ' + self.path + '/rxmd.in')
       

    def read_energy(self):
        """
        Read single-point from a completed RXMD job

        :param outfilename: Path of the stdout from the RXMD job
        :type outfilename: str
        :returns: Energy of the structure in eV
        """

        with open(self.outfile,'r') as out:

            for line in out:
                if 'MDstep:' in line:
                    energy_in_eV = float(line.strip().split()[2])*convert.kcal_mol['eV']
        return energy_in_eV
        outfile.close()


    def read_atomic_charges(outfilename):

        """
        Read atomic charge information from a completed rxmd job

        : param outfilename : Path of the file containing information about the rxmd run
        : returns: xtal object with optimized charge information
        """

        structure = xtal.AtTraj()
        snapshot = structure.create_snapshot(xtal.Snapshot)
        #outfile = open(self.path + '/DAT/000000001.xyz', 'r')

        natoms = None
        line, dummyline = None, None
        with open(self.path + '/DAT/000000001.xyz', 'r') as outfile:

            line = outfile.readline()
            natoms = int(line.strip().split()[0])

            dummyline = outfile.readline()
            structure.abc = np.array([ float(dummyline[0]), float(dummyline[1]), float(dummyline[2]) ]) 
            structure.ang = np.array([ float(dummyline[3]), float(dummyline[4]), float(dummyline[5]) ])

            for i in range(natoms):
                charges = outfile.readline()
                charges = charges.strip().split()
                atom = snapshot.create_atom(xtal.Atom)

                if charges[0] == 'C':
                    atom.element = charges[0]
                if charges[0] == 'H':
                    atom.element = charges[0]
                if charges[0] == 'O':
                    atom.element = charges[0]
                if charges[0] == 'N':
                    atom.element = charges[0]
                if charges[0] == 'S':
                    atom.element = charges[0]

                atom.charge = float(charges[4])

        return structure



# Energy conversion
eV = {'eV':1.0, 'meV':1000.0, 'kcal/mol':23.0605, 'kJ/mol':96.4853, 'Ha':0.03674932248, 'Ry':0.07349864496} # From eV
meV = {key: eV[key]/eV['meV'] for key in eV.keys()} # From meV
kcal_mol = {key: eV[key]/eV['kcal/mol'] for key in eV.keys()} # From kcal/mol
kJ_mol = {key: eV[key]/eV['kJ/mol'] for key in eV.keys()} # From kJ/mol
Ha = {key: eV[key]/eV['Ha'] for key in eV.keys()} # From Ha
Ry = {key: eV[key]/eV['Ry'] for key in eV.keys()} # From Ry

energy = {'eV':eV, 'meV':meV, 'kcal/mol':kcal_mol, 'kJ/mol':kJ_mol, 'Ha':Ha, 'Ry':Ry}
#del eV meV kcal_mol kJ_mol Ha Ry



# Frequency conversion
THz = {'THz':1.0, 'cm-1':33.35641, 'meV':4.13567, 'nm':299792.27794, 'um':299.79228, 'eV':0.00414, 'fs':1000.0, 'ps':1.0} # From THz
cm_1 = {key: THz[key]/THz['cm-1'] for key in THz.keys()} # From cm-1
meV = {key: THz[key]/THz['meV'] for key in THz.keys()} # From meV
nm = {key: THz[key]/THz['nm'] for key in THz.keys()} # From nm
um = {key: THz[key]/THz['um'] for key in THz.keys()} # From um
eV = {key: THz[key]/THz['eV'] for key in THz.keys()} # From eV
fs = {key: THz[key]/THz['fs'] for key in THz.keys()} # From fs
ps = {key: THz[key]/THz['ps'] for key in THz.keys()} # From ps

frequency = {'THz':THz, 'cm-1':cm_1, 'meV':meV, 'nm':nm, 'um':um, 'eV':eV, 'fs':fs, 'ps':ps}
#del THz cm_1 meV nm um, eV fs ps



import ezff
from ezff.interfaces import vasp, gulp
import numpy as np

# DEFINE GROUND TRUTHS
gt_relax_disp_GS = vasp.read_phonon_dispersion('ground_truths/relaxed/G_S.dat')
gt_relax_structure = vasp.read_atomic_structure('ground_truths/relaxed/POSCAR')
gt_c11 = 160.0 #GPa for C11 of MoSe2

gt_expanded_disp_GS = vasp.read_phonon_dispersion('ground_truths/expanded/G_S.dat')
gt_expanded_structure = vasp.read_atomic_structure('ground_truths/expanded/POSCAR')

gt_compressed_disp_GS = vasp.read_phonon_dispersion('ground_truths/compressed/G_S.dat')
gt_compressed_structure = vasp.read_atomic_structure('ground_truths/compressed/POSCAR')


def my_error_function(variable_values, template):

    myrank = ezff.get_pool_rank()

    # FOR THE RELAXED STRUCTURE
    path = str(myrank)+'/relaxed'
    relaxed_job = gulp.job(path=path)
    relaxed_job.structure = gt_relax_structure
    relaxed_job.forcefield = ezff.generate_forcefield(template, variable_values, FFtype = 'SW')
    relaxed_job.options['pbc'] = True
    relaxed_job.options['relax_atoms'] = True
    relaxed_job.options['relax_cell'] = True
    relaxed_job.options['phonon_dispersion'] = True
    relaxed_job.options['phonon_dispersion_from'] = '0 0 0'
    relaxed_job.options['phonon_dispersion_to'] = '0.5 0.5 0'
    # Submit job and read output
    relaxed_job.run()
    # Read output from completed GULP job and cleanup job files
    md_relax_disp_GS = relaxed_job.read_phonon_dispersion()
    md_relaxed_moduli = relaxed_job.read_elastic_moduli()
    md_relaxed_structure = relaxed_job.read_structure()
    relaxed_job.cleanup()  # FINISH RELAXED JOB


    # FOR THE COMPRESSED STRUCTURE
    path = str(myrank)+'/compressed'
    compressed_job = gulp.job(path=path)
    compressed_job.structure = gt_compressed_structure
    compressed_job.forcefield = ezff.generate_forcefield(template, variable_values, FFtype = 'SW')
    compressed_job.options['pbc'] = True
    compressed_job.options['relax_atoms'] = True
    compressed_job.options['relax_cell'] = False
    compressed_job.options['phonon_dispersion'] = True
    compressed_job.options['phonon_dispersion_from'] = '0 0 0'
    compressed_job.options['phonon_dispersion_to'] = '0.5 0.5 0'
    # Submit job and read output
    compressed_job.run()
    # Read output from completed GULP job and cleanup job files
    md_compressed_disp_GS = compressed_job.read_phonon_dispersion()
    compressed_job.cleanup()  # FINISH COMPRESSED JOB


    # FOR THE EXPANDED STRUCTURE
    path = str(myrank)+'/expanded'
    expanded_job = gulp.job(path=path)
    expanded_job.structure = gt_expanded_structure
    expanded_job.forcefield = ezff.generate_forcefield(template, variable_values, FFtype = 'SW')
    expanded_job.options['pbc'] = True
    expanded_job.options['relax_atoms'] = True
    expanded_job.options['relax_cell'] = False
    expanded_job.options['phonon_dispersion'] = True
    expanded_job.options['phonon_dispersion_from'] = '0 0 0'
    expanded_job.options['phonon_dispersion_to'] = '0.5 0.5 0'
    # Submit job and read output
    expanded_job.run()
    # Read output from completed GULP job and cleanup job files
    md_expanded_disp_GS = expanded_job.read_phonon_dispersion()
    expanded_job.cleanup()  # FINISH EXPANDED JOB

    # Compute 5 errors from the 3 GULP jobs
    error_abc, error_ang = ezff.error_lattice_constant(MD=md_relaxed_structure, GT=gt_relax_structure)
    a_lattice_error = np.linalg.norm(error_abc[0])   # Error in 'a' lattice constant
    b_lattice_error = np.linalg.norm(error_abc[1])   # Error in 'b' lattice constant

    md_c11 = md_relaxed_moduli[0][0,0] * md_relaxed_structure.box[2,2] * (2.0/13.97)  # Extracting c11 for a bulk-like layered structure from the monolayer GULP calculation
    modulus_error = np.linalg.norm(md_c11 - gt_c11)

    phon_error_relaxed = ezff.error_phonon_dispersion(MD=md_relax_disp_GS, GT=gt_relax_disp_GS, weights='acoustic')
    phon_error_expanded = ezff.error_phonon_dispersion(MD=md_expanded_disp_GS, GT=gt_expanded_disp_GS, weights='acoustic')
    phon_error_compressed = ezff.error_phonon_dispersion(MD=md_compressed_disp_GS, GT=gt_compressed_disp_GS, weights='acoustic')

    return [a_lattice_error, b_lattice_error, modulus_error, phon_error_relaxed, phon_error_compressed, phon_error_expanded]



if __name__ == '__main__':

    obj = ezff.FFParam(error_function = my_error_function, num_errors = 6)
    obj.read_variable_bounds('variable_bounds')
    obj.read_forcefield_template('template')

    pool = obj.generate_pool('mpi')

    obj.set_algorithm('randomsearch_so', population_size = 16)
    obj.parameterize(num_epochs = 5, pool = pool)
    obj.set_algorithm('nsga2_mo_platypus', population_size = 16)
    obj.parameterize(num_epochs = 25, pool = pool)



"""Interface to GULP, the General Utility Lattice Program"""
import os
import xtal
import numpy as np
import ezff
from ezff.ffio import generate_forcefield as gen_ff
from ezff.utils import convert_units as convert
import distutils
from distutils import spawn



class job:
    """
    Class representing a GULP calculation
    """
    def __init__(self, verbose=False, path='.'):
        """
        :param path: Path where the GULP job must be run from
        :type path: str

        :param verbose: Print details about the GULP job
        :type verbose: bool
        """
        if not os.path.isdir(path):
            if verbose:
                print('Path for current job is not valid . Creating a new directory...')
            os.makedirs(path)
        self.path = path

        self.scriptfile = os.path.join(os.path.abspath(path), 'in.gulp')
        self.outfile = os.path.join(os.path.abspath(path), 'out.gulp')
        self.structure = None
        self.forcefield = ''
        self.options = {
            "relax_atoms": False,
            "relax_cell": False,
            "pbc": False,
            "atomic_charges": False,
            "phonon_dispersion": None,
            "phonon_dispersion_from": None,
            "phonon_dispersion_to": None
            }
        self.verbose = verbose
        if verbose:
            print('Created a new GULP job')


    def generate_forcefield(self, template_string, parameters, FFtype = None, outfile = None):
        self.options['fftype'] = FFtype.upper()
        forcefield = gen_ff(template_string, parameters, FFtype = FFtype, outfile = outfile, MD = 'GULP')
        return forcefield


    def run(self, command = None, timeout = None):
        """
        Execute GULP job with user-defined parameters

        :param command: path to GULP executable
        :type command: str

        :param timeout: GULP job is automatically killed after ``timeout`` seconds
        :type timeout: int
        """
        if command is None:
            # Attempt to locate a `gulp` executable
            gulpexec = distutils.spawn.find_executable('gulp')
            if gulpexec is not None:
                print('Located GULP executable at ' + gulpexec)
                command = gulpexec
            else:
                print('No GULP executable specified or located')

        self._write_script_file()

        system_call_command = command + ' < ' + self.scriptfile + ' > ' + self.outfile + ' 2> ' + self.outfile + '.runerror'

        if timeout is not None:
            system_call_command = 'timeout ' + str(timeout) + ' ' + system_call_command

        if self.verbose:
            print('cd '+ self.path + ' ; ' + system_call_command)
        os.system('cd '+ self.path + ' ; ' + system_call_command)



    def _write_script_file(self):
        """
        Write-out a complete GULP script file, ``job.scriptfile``, based on job parameters
        """
        opts = self.options
        script = open(self.scriptfile,'w')
        header_line = ''
        if opts['relax_atoms']:
            header_line += 'optimise '

            if opts['relax_cell']:
                header_line += 'conp '
            else:
                header_line += 'conv '

        if header_line == '':
            header_line = 'gradient '

        if opts['phonon_dispersion'] is not None:
            header_line += 'phonon nofrequency '

        if opts['atomic_charges']:
            header_line += 'qiterative '

        header_line += 'comp property '
        script.write(header_line + '\n')

        script.write('\n')
        script.write('\n')

        # Write forcefield into script
        script.write(self.forcefield)

        script.write('\n')
        script.write('\n')

        if opts['pbc']:
            for snapshot in self.structure.snaplist:
                script.write('vectors\n')
                script.write(np.array_str(self.structure.box).replace('[','').replace(']','') + '\n')
                script.write('Fractional\n')
                for atom in snapshot.atomlist:
                    positions = atom.element.title() + ' core '
                    positions += np.array_str(atom.fract).replace('[','').replace(']','')
                    positions += ' 0.0   1.0   0.0   1 1 1 \n'
                    script.write(positions)
                script.write('\n\n\n')
        else:
            for snapshot in self.structure.snaplist:
                script.write('Cartesian\n')
                for atom in snapshot.atomlist:
                    positions = atom.element.title() + ' core '
                    positions += np.array_str(atom.cart).replace('[','').replace(']','')
                    positions += ' 0.0   1.0   0.0   1 1 1 \n'
                    script.write(positions)
                script.write('\n\n\n')
        script.write('\n')


        if opts['phonon_dispersion_from'] is not None:
            if opts['phonon_dispersion_to'] is not None:
                script.write('dispersion 1 100 \n')
                script.write(opts['phonon_dispersion_from'] + ' to ' + opts['phonon_dispersion_to']+'\n')
                script.write('output phonon ' + os.path.basename(self.outfile + '.disp'))
                script.write('\n')

        script.write('\n')

        script.close()



    def cleanup(self):
        """
        Clean-up after the completion of a GULP job. Deletes input, output and forcefields files
        """
        files_to_be_removed = [self.outfile+'.disp', self.outfile+'.dens', self.outfile, self.scriptfile, self.outfile+'.runerror']
        for file in files_to_be_removed:
            if os.path.isfile(file):
                os.remove(file)
            elif os.path.isfile(self.path+'/'+file):
                os.remove(self.path+'/'+file)



    ## OOP methods for reading output from GULP
    def read_energy(self):
        """
        Read energy from completed GULP job

        :returns: Energy of the input structure(s) in eV as a np.ndarray
        """
        return _read_energy(self.outfile)

    def read_elastic_moduli(self):
        """
        Read elastic modulus matrix from a completed GULP job

        :returns: 6x6 Elastic modulus matrix in GPa for each input structure, as a list
        """
        return _read_elastic_moduli(self.outfile)

    def read_phonon_dispersion(self, units='cm-1'):
        """
        Read phonon dispersion from a complete GULP job

        :returns: 2D np.array containing the phonon dispersion in THz
        """
        return _read_phonon_dispersion(self.outfile+'.disp', units=units)

    def read_atomic_charges(self):
        """
        Read atomic charge information from a completed GULP job file

        :returns: xtal.AtTraj object with optimized charge information
        """
        return _read_atomic_charges(self.outfile)

    def read_structure(self):
        """
        Read converged structure (cell and atomic positions) from the MD job

        :returns: xtal.AtTraj object with (optimized) individual structures as separate snapshots
        """
        return _read_structure(self.outfile, relax_cell=self.options['relax_cell'], initial_box=self.structure.box)



def _read_elastic_moduli(outfilename):
    """
    Read elastic modulus matrix from a completed GULP job

    :param outfilename: Path of the stdout from the GULP job
    :type outfilename: str
    :returns: 6x6 Elastic modulus matrix in GPa
    """
    outfile = open(outfilename,'r')
    moduli_array = []
    while True:
        oneline = outfile.readline()
        if not oneline: # break at EOF
            break
        if 'Elastic Constant Matrix' in oneline:
            moduli = np.zeros((6,6))
            dummyline = outfile.readline()
            dummyline = outfile.readline()
            dummyline = outfile.readline()
            dummyline = outfile.readline()
            for i in range(6):
                modline = outfile.readline().strip()
                e1, e2, e3, e4, e5, e6 = modline[3:13], modline[13:23], modline[23:33], modline[33:43], modline[43:53], modline[53:63]
                modarray = [e1,e2,e3,e4,e5,e6]
                float_modarray = []
                # Handle errors
                for element in modarray:
                    if element[0] == "*":
                        float_modarray.append(0.0)
                    else:
                        float_modarray.append(float(element))
                moduli[i,:] = float_modarray
            moduli_array.append(moduli)
    outfile.close()
    return moduli_array



def _read_energy(outfilename):
    """
    Read single-point from a completed GULP job

    :param outfilename: Path of the stdout from the GULP job
    :type outfilename: str
    :returns: Energy of the structure in eV
    """
    energy_in_eV = []
    outfile = open(outfilename, 'r')
    for line in outfile:
        if 'Total lattice energy' in line:
            if line.strip().split()[-1] == 'eV':
                energy_in_eV.append(float(line.strip().split()[-2]))
    outfile.close()
    return np.array(energy_in_eV)



def _read_phonon_dispersion(phonon_dispersion_file, units='cm-1'):
    """
    Read phonon dispersion from a complete GULP job

    :param phonon_dispersion_file: Path of file containing phonon dispersion from the GULP job
    :type phonon_dispersion_file: str
    :returns: 2D np.array containing the phonon dispersion in THz
    """
    pbs = []
    freq_conversion = {'cm-1': 0.0299792453684314, 'THz': 1, 'eV': 241.79893, 'meV': 0.24180}
    dispersion = open(phonon_dispersion_file, 'r')
    for line in dispersion:
        if not line.strip().startswith('#'):
            str_index, str_freq = line[:4], line[4:]
            if not str_freq[0] == '*':
                pbs.append(float(str_freq))
            else:
                pbs.append(0.0)
    dispersion.close()
    num_bands = int(len(pbs)/100)
    pbs = np.array(pbs).reshape((100,num_bands)).T
    pbs *= convert.frequency[units]['THz']
    return pbs



def _read_atomic_charges(outfilename):
    """
    Read atomic charge information from a completed GULP job file

    :param outfilename: Filename of the GULP output file
    :type outfilename: str
    :returns: xtal object with optimized charge information
    """
    structure = xtal.AtTraj()
    structure.box = np.zeros((3,3))
    outfile = open(outfilename, 'r')

    while True:
        oneline = outfile.readline()
        if not oneline:  # EOF check
            break
        if 'Output for configuration' in oneline:
            snapshot = structure.create_snapshot(xtal.Snapshot)
        if 'Final charges from ReaxFF' in oneline:
            snapshot.atomlist = []
            dummyline = outfile.readline()
            dummyline = outfile.readline()
            dummyline = outfile.readline()
            dummyline = outfile.readline()
            while True:
                charges = outfile.readline().strip().split()
                if charges[0][0] == '-':
                    break
                else:
                    atom = snapshot.create_atom(xtal.Atom)
                    atom.charge = float(charges[-1])
    return structure



def _read_structure(outfilename, relax_cell=True, initial_box=None):
    """
    Read converged structure (cell and atomic positions) from the MD job

    :param outfilename: Path of file containing stdout of the GULP job
    :type outfilename: str

    :param relax_cell: Flag to identify if simulation cell was relaxed during the MD job
    :type relax_cell: boolean

    :param initial_box: Initial simulation cell used for the MD job
    :type initial_box: 3X3 np.ndarray

    :returns: xtal.AtTraj object with (optimized) individual structures as separate snapshots
    """
    relaxed = xtal.AtTraj()
    relaxed.box = np.zeros((3,3))
    if (not relax_cell) and (initial_box is not None):
        relaxed.box = initial_box

    # Read number of atoms
    outfile = open(outfilename, 'r')
    for line in outfile:
        if 'Number of irreducible atoms/shells' in line.strip():
            snapshot = relaxed.create_snapshot(xtal.Snapshot)
            num_atoms = int(line.strip().split()[-1])
            for atomID in range(num_atoms):
                atom = snapshot.create_atom(xtal.Atom)
                atom.cart = np.array([0.0,0.0,0.0])
                atom.fract = np.array([0.0,0.0,0.0])
    outfile.close()

    snapID = 0
    convert_to_cart = True
    outfile = open(outfilename, 'r')

    while True:
        oneline = outfile.readline()

        if not oneline: # break for EOF
            break

        if 'Comparison of initial and final' in oneline:
            dummyline = outfile.readline()
            dummyline = outfile.readline()
            dummyline = outfile.readline()
            dummyline = outfile.readline()
            while True:
                data = outfile.readline().strip().split()
                if data[0].isdigit():
                    atomID = int(data[0])-1
                    if data[1] == 'x':
                        axisID = 0
                    elif data[1] == 'y':
                        axisID = 1
                    else:
                        axisID = 2

                    if data[5] == 'Cartesian':
                        relaxed.snaplist[snapID].atomlist[atomID].cart[axisID] = float(data[3])
                        convert_to_cart = False
                    elif data[5] == 'Fractional':
                        relaxed.snaplist[snapID].atomlist[atomID].fract[axisID] = float(data[3])
                elif data[0][0] == '-':
                    break

            snapID += 1
    outfile.close()

    if convert_to_cart:
        if relax_cell:
            outfile = open(outfilename, 'r')
            snapID = 0
            while True:
                oneline = outfile.readline()

                if not oneline: # break for EOF
                    break

                if 'Final Cartesian lattice vectors (Angstroms)' in oneline:
                    dummyline = outfile.readline()
                    for i in range(3):
                        relaxed.box[i][0:3] = list(map(float,outfile.readline().strip().split()))
                    relaxed.make_dircar_matrices()
                    relaxed.snaplist[snapID].dirtocar()
                    snapID += 1
            outfile.close()
        else:
            relaxed.make_dircar_matrices()
            relaxed.dirtocar()

    return relaxed



"""This module provide methods to handle reading and writing forcefields"""
import numpy as np
import time
from ezff.utils.reaxff import reax_forcefield

def read_variable_bounds(filename, verbose=False):
    """Read permissible lower and upper bounds for decision variables used in forcefields optimization

    :param filename: Name of text file listing bounds for each decision variable that must be optimized
    :type filename: str

    :param verbose: Print all variables read-in
    :type verbose: bool
    """
    variable_bounds = {}
    while True: # Force-read the parameter bounds file. This will loop until something is read-in. This is required if multiple ranks access the same file at the same time
        time.sleep(np.random.rand())
        with open(filename, 'r') as variable_bounds_file:
            for line in variable_bounds_file:
                items = line.strip().split()
                key, values = items[0], items[1:]
                if key[0] == '_':
                    variable_bounds[key] = list(map(int, values))
                else:
                    variable_bounds[key] = list(map(float, values))
        if not variable_bounds == {}:
            break

    if verbose:
        allkeys = ''
        for key in variable_bounds:
            allkeys += str(key) + ', '
        print('Keys: ' + allkeys[:-2] + ' read from ' + filename)

    return variable_bounds



def read_forcefield_template(template_filename):
    """Read-in the forcefield template. The template is constructed from a functional forcefield file by replacing all optimizable numerical values with variable names enclosed within dual angled brackets << and >>.

    :param template_filename: Name of the forcefield template file to be read-in
    :type template_filename: str
    """
    while True: # Force-read the template forcefield. This will loop until something is read-in. This is required if multiple ranks read the same file at the same time
        time.sleep(np.random.rand())
        with open(template_filename) as forcefield_template_file:
            template_string = forcefield_template_file.read()
        if len(template_string) > 0:
            break

    return template_string



def generate_forcefield(template_string, parameters, FFtype = None, outfile = None, MD = 'GULP'):
    """Generate a new forcefield from the template by replacing variables with numerical values

    :param template_string: Text of the forcefield template
    :type template_string: str

    :param parameters: Numerical value of all decision variables in the form of variable:value pairs
    :type parameters: dict

    :param FFtype: Type of forcefield being optimized. (e.g. reaxff, sw, lj, etc.)
    :type FFtype: string

    :param outfile: Optional filename to write out the forcefield
    :type outfile: string

    :param MD: MD Engine used for parameterization
    :type MD: string
    """
    replaced_keys = ''
    for key, ranges in parameters.items():
        pattern = '<<' + key + '>>'
        while pattern in template_string:
            template_string = template_string.replace(pattern, '%12.6f' % ranges)

    # Special handling for certain forcefield types
    if FFtype is not None:
        if (FFtype.strip().upper() == 'REAX') or (FFtype.strip().upper() == 'REAXFF'):
            if MD == 'GULP':
                template_string = reax_forcefield(filestring=template_string).write_gulp_library()
            elif MD == 'LAMMPS':
                template_string = reax_forcefield(filestring=template_string).write_formatted_forcefields()

    if outfile is not None:
        with open(outfile, 'w') as new_forcefield:
            new_forcefield.write(template_string)
        return
    else:
        return template_string



import ezff
from ezff.interfaces import gulp, vasp, lammps
import numpy as np

# DEFINE GROUND TRUTHS
gt_bulk_modulus = 2.65 #GPa
gt_structure = vasp.read_atomic_structure('ground_truths/POSCAR')

def my_error_function(variable_values, template):

    myrank = ezff.get_pool_rank()

    # Configure LAMMPS job.
    path = str(myrank)
    md_job = lammps.job(path=path)
    md_job.structure = gt_structure
    md_job.forcefield = md_job.generate_forcefield(template, variable_values, FFtype='LJ')
    md_job.options['pbc'] = True
    md_job.options['relax_atoms'] = True
    md_job.options['relax_cell'] = True

    # Submit job and read output
    md_job.run_elastic()

    # Calculate errors in lattice constant and elastic modulus
    moduli = md_job.read_elastic_moduli()                                                # 6X6 elastic modulus tensor inside a list of length 1 (for a single input structure)
    try:
        md_bulk_modulus = np.mean([moduli[0][i,i] for i in range(3)])                        # Bulk modulus is sum of diagonal elements in moduli[0]
        bulk_modulus_error = np.linalg.norm(md_bulk_modulus - gt_bulk_modulus)               # Error is the L2 norm of deviation from the ground truth value
    except:
        bulk_modulus_error = np.nan

    md_structure = md_job.read_structure()                                               # Read relaxed structure after optimization
    error_abc, error_ang = ezff.error_lattice_constant(MD=md_structure, GT=gt_structure) # error_abc = error in lattice constants, error_ang = error in lattice angles
    latt_error_1 = np.linalg.norm(error_abc[0])                                            # error in 'a' lattice constant
    latt_error_2 = np.linalg.norm(error_abc[1])                                            # error in 'a' lattice constant

    md_job.cleanup()

    return [latt_error_1+latt_error_2, bulk_modulus_error]


if __name__ == '__main__':

    obj = ezff.FFParam(error_function = my_error_function, num_errors = 2)
    obj.read_variable_bounds('variable_bounds')
    obj.read_forcefield_template('template')
    obj.set_algorithm('randomsearch_so', population_size = 50)
    obj.parameterize(num_epochs = 10)
    obj.set_algorithm('NSGA2_MO_PLATYPUS', population_size = 50)
    obj.parameterize(num_epochs = 10)



import ezff
from ezff.interfaces import lammps, qchem
from ezff.utils.reaxff import reax_forcefield
import time

# Define ground truths
gt_gs = qchem.read_structure('ground_truths/optCHOSx.out')
gt_gs_energy = qchem.read_energy('ground_truths/optCHOSx.out')
gt_freq_scan = qchem.read_structure('ground_truths/frequency_length_scan/CHOSx.out')
gt_freq_scan_energy = qchem.read_energy('ground_truths/frequency_length_scan/CHOSx.out')
gt_full_scan = qchem.read_structure(['ground_truths/dissociation_length_scan/CHOSx.run1.out', 'ground_truths/dissociation_length_scan/CHOSx.run2.out', 'ground_truths/dissociation_length_scan/CHOSx.run3.out'])
gt_full_scan_energy = qchem.read_energy(['ground_truths/dissociation_length_scan/CHOSx.run1.out', 'ground_truths/dissociation_length_scan/CHOSx.run2.out', 'ground_truths/dissociation_length_scan/CHOSx.run3.out'])

def my_error_function(variable_values, template):

    myrank = ezff.get_pool_rank()

    path = str(myrank)+'/relaxed'

    # Calculate Ground State
    md_gs_job = lammps.job(path = path)
    md_gs_job.structure = gt_gs
    md_gs_job.forcefield = md_gs_job.generate_forcefield(template, variable_values, FFtype = 'reaxff')
    md_gs_job.options['pbc'] = False
    md_gs_job.options['relax_atoms'] = False
    md_gs_job.options['relax_cell'] = False
    # Run GULP calculation
    md_gs_job.run()
    # Read output from completed GULP job and clean-up
    md_gs_energy = md_gs_job.read_energy()
    #md_gs_job.cleanup()

    # Calculate PES Scan for frequency error
    md_freq_scan_job = lammps.job(path = path)
    md_freq_scan_job.structure = gt_freq_scan
    md_freq_scan_job.forcefield = md_freq_scan_job.generate_forcefield(template, variable_values, FFtype = 'reaxff')
    md_freq_scan_job.options['pbc'] = False
    md_freq_scan_job.options['relax_atoms'] = False
    md_freq_scan_job.options['relax_cell'] = False
    # Run GULP calculation
    md_freq_scan_job.run()
    # Read output from completed GULP job and clean-up
    md_freq_scan_energy = md_freq_scan_job.read_energy()
    #md_freq_scan_job.cleanup()

    # Calculate PES Scan for dissociation error
    md_full_scan_job = lammps.job(path = path)
    md_full_scan_job.structure = gt_full_scan
    md_full_scan_job.forcefield = md_full_scan_job.generate_forcefield(template, variable_values, FFtype = 'reaxff')
    md_full_scan_job.options['pbc'] = False
    md_full_scan_job.options['relax_atoms'] = False
    md_full_scan_job.options['relax_cell'] = False
    # Run GULP calculation
    md_full_scan_job.run()
    # Read output from completed GULP job and clean-up
    md_full_scan_energy = md_full_scan_job.read_energy()
    #md_full_scan_job.cleanup()


    # Calculate error
    freq_error = ezff.error_energy(md_freq_scan_energy-md_gs_energy, gt_freq_scan_energy-gt_gs_energy, weights = 'minima')
    dissociation_error = ezff.error_energy(md_full_scan_energy-md_gs_energy, gt_full_scan_energy-gt_gs_energy, weights = 'dissociation')
    return [freq_error, dissociation_error]


# Generate forcefield template and variable ranges
FF = reax_forcefield('ffield')
FF.make_template_twobody('S','C',double_bond=True)
FF.make_template_threebody('S','C','S')
FF.make_template_fourbody('S','C','S','S')
FF.generate_templates()

time.sleep(1.0)

if __name__ == '__main__':

    obj = ezff.FFParam(error_function = my_error_function, num_errors = 2)
    obj.read_variable_bounds('param_ranges')
    obj.read_forcefield_template('ff.template.generated')

    pool = obj.generate_pool('multi')

    obj.set_algorithm('randomsearch_so', population_size = 9)
    obj.parameterize(num_epochs = 5, pool = pool)
    obj.set_algorithm('NSGA2_MO_PLATYPUS', population_size = 9)
    obj.parameterize(num_epochs = 5, pool = pool)



"""This module provide general functions for EZFF"""

# General imports
import os
import sys
import numpy as np
import math
import random
from datetime import datetime
import xtal
from functools import partial

# EZFF imports
from .ffio import *
from .errors import *

# Optimizer imports
import nevergrad as ng
import mobopt
from pymoo.core.problem import Problem as pymoo_Problem
from pymoo.core.individual import Individual as pymoo_Individual
from pymoo.core.population import Population as pymoo_Population
from pymoo.algorithms.moo.nsga2 import NSGA2 as pymoo_NSGA2
from pymoo.algorithms.moo.nsga3 import NSGA3 as pymoo_NSGA3
from pymoo.algorithms.moo.unsga3 import UNSGA3 as pymoo_UNSGA3
from pymoo.algorithms.moo.ctaea import CTAEA as pymoo_CTAEA
from pymoo.algorithms.moo.sms import SMSEMOA as pymoo_SMSEMOA
from pymoo.util.ref_dirs import get_reference_directions as pymoo_get_reference_directions
from pymoo.algorithms.moo.rvea import RVEA as pymoo_RVEA
from pymoo.termination.max_eval import MaximumFunctionCallTermination
from pymoo.algorithms.soo.nonconvex.es import ES as pymoo_ES
from pymoo.algorithms.soo.nonconvex.nelder import NelderMead as pymoo_NelderMead
from pymoo.algorithms.soo.nonconvex.cmaes import CMAES as pymoo_CMAES
import platypus

# Parallel processing imports
from schwimmbad import MultiPool as sch_MultiPool
from schwimmbad import MPIPool as sch_MPIPool
from mpi4py import MPI
import multiprocessing



__version__ = '1.0.0' # Update setup.py if version changes


class FFParam(object):
    """
    Class for EZFF Forcefield Parametrization
    """

    def __init__(self, error_function=None, num_errors=None):
        """
        :param num_errors: Number of errors to be minimized for forcefield optimization
        :type num_errors: int

        :param error_function: User-defined function that takes-in a dictionary of variable-value pairs and outputs a list of computed errors
        :type error_function: function
        """
        self.error_function = error_function
        self.num_errors = num_errors
        self.relative_weights = np.array([1.0 for i in range(num_errors)])
        self.variables = []
        self.errors = []
        self.total_epochs = 0

    def read_variable_bounds(self, filename):
        """Read permissible lower and upper bounds for decision variables used in forcefields optimization

        :param filename: Name of text file listing bounds for each decision variable that must be optimized
        :type filename: str
        """
        self.variable_bounds = ffio.read_variable_bounds(filename)
        self.num_variables = len(self.variable_bounds.keys())
        self.variable_names = [key for key in self.variable_bounds.keys()]


    def read_forcefield_template(self, template_filename):
        """Read-in the forcefield template. The template is constructed from a functional forcefield file by replacing all optimizable numerical values with variable names enclosed within dual angled brackets << and >>.

        :param template_filename: Name of the forcefield template file to be read-in
        :type template_filename: str
        """
        self.forcefield_template = ffio.read_forcefield_template(template_filename)

    def set_algorithm(self, algo_string, population_size = None):
        """
        Set optimization algorithm. Initialize interfaces to external optimizers and return the algorithm object

        :param algo_string: Type of algorithm to parameterize the forcefield
        :type algo_string: str

        :param population_size: Number of candidate forcefields evaluated every epoch
        :type algo_string: int
        """

        self.population_size = population_size
        self.algo_string = algo_string

        ng_algos = ['NGOPT_SO', 'TWOPOINTSDE_SO','PORTFOLIODISCRETEONEPLUSONE_SO','ONEPLUSONE_SO','CMA_SO','TBPSA_SO', 'PSO_SO', 'SCRHAMMERSLEYSEARCHPLUSMIDDLEPOINT_SO', 'RANDOMSEARCH_SO']
        mobopt_algos = ['MOBO']
        pymoo_algos = ['NSGA2_MO_PYMOO', 'NSGA3_MO_PYMOO', 'UNSGA3_MO_PYMOO', 'CTAEA_MO_PYMOO', 'SMSEMOA_MO_PYMOO', 'RVEA_MO_PYMOO', 'ES_SO_PYMOO', 'NELDERMEAD_SO_PYMOO', 'CMAES_SO_PYMOO']
        platypus_algos = ['NSGA2_MO_PLATYPUS', 'NSGA3_MO_PLATYPUS', 'GDE3_MO_PLATYPUS']

        if algo_string.upper() in ng_algos:
            self.algo_framework = 'nevergrad'
            ng_variable_dict = ng.p.Dict()
            for variable in self.variable_bounds.keys():
                ng_variable_dict[variable] = ng.p.Scalar(lower = self.variable_bounds[variable][0], upper = self.variable_bounds[variable][1])

            if algo_string.upper() == 'NGOPT_SO':
                self.algorithm = ng.optimizers.NGOpt(parametrization=ng_variable_dict, budget=self.population_size, num_workers=2)
            elif algo_string.upper() == 'TWOPOINTSDE_SO':
                self.algorithm = ng.optimizers.TwoPointsDE(parametrization=ng_variable_dict, budget=self.population_size, num_workers=2)
            elif algo_string.upper() == 'PORTFOLIODISCRETEONEPLUSONE_SO':
                self.algorithm = ng.optimizers.PortfolioDiscreteOnePlusOne_SO(parametrization=ng_variable_dict, budget=self.population_size, num_workers=2)
            elif algo_string.upper() == 'ONEPLUSONE_SO':
                self.algorithm = ng.optimizers.OnePlusOne(parametrization=ng_variable_dict, budget=self.population_size, num_workers=2)
            elif algo_string.upper() == 'CMA_SO':
                self.algorithm = ng.optimizers.CMA(parametrization=ng_variable_dict, budget=self.population_size, num_workers=2)
            elif algo_string.upper() == 'TBPSA_SO':
                self.algorithm = ng.optimizers.TBPSA(parametrization=ng_variable_dict, budget=self.population_size, num_workers=2)
            elif algo_string.upper() == 'PSO_SO':
                self.algorithm = ng.optimizers.PSO(parametrization=ng_variable_dict, budget=self.population_size, num_workers=2)
            elif algo_string.upper() == 'SCRHAMMERSLEYSEARCHPLUSMIDDLEPOINT_SO':
                self.algorithm = ng.optimizers.ScrHammersleySearchPlusMiddlePoint(parametrization=ng_variable_dict, budget=self.population_size, num_workers=2)
            elif algo_string.upper() == 'RANDOMSEARCH_SO':
                self.algorithm = ng.optimizers.RandomSearch(parametrization=ng_variable_dict, budget=self.population_size, num_workers=2)

        elif algo_string.upper() in mobopt_algos:
            self.algo_framework = 'mobopt'
            var_mins = []
            var_maxs = []
            for variable in self.variable_bounds.keys():
                var_mins.append(self.variable_bounds[variable][0])
                var_maxs.append(self.variable_bounds[variable][1])
            self.mobopt_variable_bounds = np.vstack((var_mins, var_maxs)).T
            self.algorithm = mobopt.MOBayesianOpt(target = self.error_function, NObj = self.num_errors, pbounds = self.mobopt_variable_bounds)

        elif algo_string.upper() in pymoo_algos:
            self.algo_framework = 'pymoo'
            var_mins = []
            var_maxs = []
            for variable in self.variable_bounds.keys():
                var_mins.append(self.variable_bounds[variable][0])
                var_maxs.append(self.variable_bounds[variable][1])
            self.pymoo_problem = pymoo_Problem(n_var = self.num_variables, n_obj = self.num_errors, n_constr = 0, xl = var_mins, xu = var_maxs)

            self.normalize_errors()
            initial_population = []
            for varid, var in enumerate(self.variables):
                evaledsoln = pymoo_Individual()
                evaledsoln._X = var
                evaledsoln._F = self.normalized_errors[varid]
                initial_population.append(evaledsoln)

            if algo_string.upper() == 'NSGA2_MO_PYMOO':
                if initial_population == []:
                    self.algorithm = pymoo_NSGA2(self.population_size)
                else:
                    initial_population = pymoo_Population(initial_population)
                    self.algorithm = pymoo_NSGA2(self.population_size, sampling = initial_population)
                self.algorithm.setup(self.pymoo_problem, seed = np.random.randint(100000), verbose = False)

            elif algo_string.upper() == 'NSGA3_MO_PYMOO':
                # Identify number of reference points
                min_points = [math.comb(self.num_errors + ref_pts - 1, ref_pts) for ref_pts in range(25)]
                num_reference_points = np.sum(np.array(min_points) < self.population_size)
                reference_directions = pymoo_get_reference_directions("das-dennis", self.num_errors, n_partitions=num_reference_points)
                if initial_population == []:
                    self.algorithm = pymoo_NSGA3(pop_size=self.population_size, ref_dirs=reference_directions)
                else:
                    initial_population = pymoo_Population(initial_population)
                    self.algorithm = pymoo_NSGA3(pop_size=self.population_size, ref_dirs=reference_directions, sampling = initial_population)
                self.algorithm.setup(self.pymoo_problem, seed = np.random.randint(100000), verbose = False)

            elif algo_string.upper() == 'UNSGA3_MO_PYMOO':
                # Identify number of reference points
                min_points = [math.comb(self.num_errors + ref_pts - 1, ref_pts) for ref_pts in range(25)]
                num_reference_points = np.sum(np.array(min_points) < self.population_size)
                reference_directions = pymoo_get_reference_directions("das-dennis", self.num_errors, n_partitions=num_reference_points)
                if initial_population == []:
                    self.algorithm = pymoo_UNSGA3(pop_size=self.population_size, ref_dirs=reference_directions)
                else:
                    initial_population = pymoo_Population(initial_population)
                    self.algorithm = pymoo_UNSGA3(pop_size=self.population_size, ref_dirs=reference_directions, sampling = initial_population)
                self.algorithm.setup(self.pymoo_problem, seed = np.random.randint(100000), verbose = False)

            elif algo_string.upper() == 'CTAEA_MO_PYMOO':
                # Identify number of reference points
                min_points = [math.comb(self.num_errors + ref_pts - 1, ref_pts) for ref_pts in range(25)]
                num_reference_points = np.sum(np.array(min_points) < self.population_size)
                reference_directions = pymoo_get_reference_directions("das-dennis", self.num_errors, n_partitions=num_reference_points)
                if initial_population == []:
                    self.algorithm = pymoo_CTAEA(ref_dirs=reference_directions)
                else:
                    initial_population = pymoo_Population(initial_population)
                    self.algorithm = pymoo_CTAEA(ref_dirs=reference_directions, sampling = initial_population)
                self.algorithm.setup(self.pymoo_problem, seed = np.random.randint(100000), verbose = False)

            elif algo_string.upper() == 'SMSEMOA_MO_PYMOO':
                if initial_population == []:
                    self.algorithm = pymoo_SMSEMOA(self.population_size)
                else:
                    initial_population = pymoo_Population(initial_population)
                    self.algorithm = pymoo_SMSEMOA(self.population_size, sampling = initial_population)
                self.algorithm.setup(self.pymoo_problem, seed = np.random.randint(100000), verbose = False)

            elif algo_string.upper() == 'RVEA_MO_PYMOO':
                # Identify number of reference points
                min_points = [math.comb(self.num_errors + ref_pts - 1, ref_pts) for ref_pts in range(25)]
                num_reference_points = np.sum(np.array(min_points) < self.population_size)
                reference_directions = pymoo_get_reference_directions("das-dennis", self.num_errors, n_partitions=num_reference_points)
                if initial_population == []:
                    self.algorithm = pymoo_RVEA(ref_dirs=reference_directions, pop_size = self.population_size)
                    termination = MaximumFunctionCallTermination(5000)
                    self.algorithm.termination = termination
                else:
                    initial_population = pymoo_Population(initial_population)
                    self.algorithm = pymoo_RVEA(ref_dirs=reference_directions, sampling = initial_population, pop_size = self.population_size)
                    termination = MaximumFunctionCallTermination(5000)
                    self.algorithm.termination = termination
                self.algorithm.setup(self.pymoo_problem, seed = np.random.randint(100000), verbose = False)

            elif algo_string.upper() == 'ES_SO_PYMOO':
                if initial_population == []:
                    self.algorithm = pymoo_ES(n_offspring = self.population_size, pop_size = self.population_size)
                else:
                    initial_population = pymoo_Population(initial_population)
                    self.algorithm = pymoo_ES(n_offspring = self.population_size, pop_size = self.population_size, sampling = initial_population)
                self.algorithm.setup(self.pymoo_problem, seed = np.random.randint(100000), verbose = False)

            elif algo_string.upper() == 'NELDERMEAD_SO_PYMOO':
                if initial_population == []:
                    self.algorithm = pymoo_NelderMead()
                else:
                    initial_population = pymoo_Population(initial_population)
                    self.algorithm = pymoo_NelderMead()
                    self.algorithm.pop = initial_population
                self.algorithm.setup(self.pymoo_problem, seed = np.random.randint(100000), verbose = False)

            elif algo_string.upper() == 'CMAES_SO_PYMOO':
                x0_search = np.mean(self.variables, axis=0)
                if initial_population == []:
                    self.algorithm = pymoo_CMAES(x0 = x0_search)
                else:
                    initial_population = pymoo_Population(initial_population)
                    self.algorithm = pymoo_CMAES(x0 = x0_search)
                    self.algorithm.pop = initial_population
                self.algorithm.setup(self.pymoo_problem, seed = np.random.randint(100000), verbose = False)

        elif algo_string.upper() in platypus_algos:
            self.algo_framework = 'platypus'
            var_mins = []
            var_maxs = []
            for variable in self.variable_bounds.keys():
                var_mins.append(self.variable_bounds[variable][0])
                var_maxs.append(self.variable_bounds[variable][1])
            self.platypus_problem = platypus.Problem(self.num_variables, self.num_errors)
            self.platypus_problem.types = [platypus.Real(var_mins[i], var_maxs[i]) for i in range(self.num_variables)]

            self.normalize_errors()
            initial_population = []
            for varid, var in enumerate(self.variables):
                evaledsoln = platypus.Solution(self.platypus_problem)
                evaledsoln.variables[:] = list(var)
                evaledsoln.objectives[:] = list(self.normalized_errors[varid])
                initial_population.append(evaledsoln)

            if algo_string.upper() == 'NSGA2_MO_PLATYPUS':
                self.algorithm = platypus.NSGAII(self.platypus_problem, self.population_size)
                self.algorithm.variator = platypus.default_variator(self.platypus_problem)
                if len(initial_population) > 0:
                    self.algorithm.population = initial_population

            elif algo_string.upper() == 'NSGA3_MO_PLATYPUS':
                # Identify number of reference points
                min_points = [math.comb(self.num_errors + ref_pts - 1, ref_pts) for ref_pts in range(200)]
                num_reference_points = np.sum(np.array(min_points) < self.population_size)
                self.algorithm = platypus.NSGAIII(self.platypus_problem, divisions_outer = num_reference_points)
                self.algorithm.variator = platypus.default_variator(self.platypus_problem)
                if len(initial_population) > 0:
                    self.algorithm.population = initial_population

            if algo_string.upper() == 'GDE3_MO_PLATYPUS':
                self.algorithm = platypus.GDE3(self.platypus_problem, population_size = self.population_size)
                if len(initial_population) > 0:
                    self.algorithm.population = initial_population



    def ask(self):
        """
        Ask the optimization algorithm the next candidate variables for evaluation
        """
        new_variables = []
        if self.algo_framework == 'nevergrad':
            for i in range(self.population_size):
                newvar = self.algorithm.ask()
                new_var_as_list = np.array([newvar.value[self.variable_names[i]] for i in range(len(self.variable_names))])
                new_variables.append(new_var_as_list)
            return new_variables

        elif self.algo_framework == 'mobopt':
            q = 0.5
            prob = 0.1

            for pointID in range(len(self.variables)):
                self.algorithm.space.add_observation(np.array(self.variables[pointID]), np.array(self.normalized_errors[pointID]))

            for i in range(self.num_errors):
                yy = self.algorithm.space.f[:, i]
                self.algorithm.GP[i].fit(self.algorithm.space.x, yy)

            pop, logbook, front = mobopt._NSGA2.NSGAII(self.algorithm.NObj,
                                                       self.algorithm._MOBayesianOpt__ObjectiveGP,
                                                       self.algorithm.pbounds,
                                                       MU=self.population_size*2)


            Population = np.asarray(pop)
            IndexF, FatorF = self.algorithm._MOBayesianOpt__LargestOfLeast(front, self.algorithm.space.f)
            IndexPop, FatorPop = self.algorithm._MOBayesianOpt__LargestOfLeast(Population, self.algorithm.space.x)

            Fator = q * FatorF + (1-q) * FatorPop
            sorted_ids = np.argsort(Fator)


            for i in range(self.population_size):
                Index_try = int(np.argwhere(sorted_ids == np.max(sorted_ids)-i))
                self.algorithm.x_try = Population[Index_try]

                if self.algorithm.space.RS.uniform() < prob:
                    if self.algorithm.NParam > 1:
                        ii = self.algorithm.space.RS.randint(low=0, high=self.algorithm.NParam - 1)
                    else:
                        ii = 0

                    self.algorithm.x_try[ii] = self.algorithm.space.RS.uniform(low=self.algorithm.pbounds[ii][0],high=self.algorithm.pbounds[ii][1])

                new_variables.append(self.algorithm.x_try)

            return new_variables


        elif self.algo_framework == 'pymoo':
            while (len(new_variables) < self.population_size):
                newXs = self.algorithm.ask()

                for newX in newXs:
                    new_variables.append(newX.X)

            return new_variables


        elif self.algo_framework == 'platypus':
            if self.algo_string.upper() == 'NSGA2_MO_PLATYPUS':
                platypus.nondominated_sort(self.algorithm.population)
                self.algorithm.population = platypus.nondominated_truncate(self.algorithm.population, self.population_size)
                for i in range(self.population_size):
                    parents = self.algorithm.selector.select(self.algorithm.variator.arity, self.algorithm.population)
                    single_offspring = self.algorithm.variator.evolve(parents)
                    new_variables.append(single_offspring[0].variables[:])
                    new_variables.append(single_offspring[1].variables[:])
                new_variables = random.sample(new_variables, self.population_size)
            elif self.algo_string.upper() == 'NSGA3_MO_PLATYPUS':
                platypus.nondominated_sort(self.algorithm.population)
                self.algorithm.population = self.algorithm._reference_point_truncate(self.algorithm.population, self.population_size)
                for i in range(self.population_size):
                    parents = self.algorithm.selector.select(self.algorithm.variator.arity, self.algorithm.population)
                    single_offspring = self.algorithm.variator.evolve(parents)
                    new_variables.append(single_offspring[0].variables[:])
                    new_variables.append(single_offspring[1].variables[:])
                new_variables = random.sample(new_variables, self.population_size)
            elif self.algo_string.upper() == 'GDE3_MO_PLATYPUS':
                self.algorithm.population = self.algorithm.survival(self.algorithm.population)
                for i in range(self.population_size):
                    parents = self.algorithm.select(i, self.algorithm.variator.arity)
                    single_offspring = self.algorithm.variator.evolve(parents)
                    new_variables.append(single_offspring[0].variables[:])
                #     new_variables.append(single_offspring[1].variables[:])
                # new_variables = random.sample(list(set(new_variables)), self.population_size)

            return new_variables



    def parameterize(self, num_epochs = None, pool = None):
        """
        The optimize function provides a uniform wrapper to solve the EZFF problem using the algorithm(s) provided.

        :param num_epochs: Number of epochs to perform the optimization for. If multiple algorithms are specified, one iteration value should be provided for each algorithm
        :type num_epochs: int

        :param pool: Multiprocessing or MPI Pool for forcefield parameterization
        :type pool: Multiprocessing or MPI Pool object
        """
        self.pool = pool
        if self.algo_framework == 'nevergrad':
            for epoch in range(num_epochs):

                self.total_epochs += 1
                self.normalize_errors()

                self.set_algorithm(algo_string = self.algo_string, population_size = self.population_size)

                for computed_id, computed in enumerate(self.variables):
                    computed_value = {k:v for (k,v) in zip(self.variable_names,computed)}
                    self.algorithm.suggest(computed_value)
                    asked_suggestion = self.algorithm.ask()
                    self.algorithm.tell(asked_suggestion, self.normalized_errors[computed_id])

                new_variables = self.ask()
                new_errors = []

                if self.pool is None:
                    for variable in new_variables:
                        variable_dict = dict(zip(self.variable_names, variable))
                        error = self.error_function(variable_dict, self.forcefield_template)
                        new_errors.append(error)
                else:
                    if self.pool_type == 'mpi':
                        if not self.pool.is_master():
                            pool.wait()

                    variable_dict_list = [dict(zip(self.variable_names, variable)) for variable in new_variables]
                    new_errors = self.pool.map(partial(self.error_function, template = self.forcefield_template), variable_dict_list)

                for variable_id, variable in enumerate(new_variables):
                    self.variables.append(variable)
                    self.errors.append(new_errors[variable_id])

                self._write_out_forcefields()


        elif self.algo_framework == 'mobopt':
            for epoch in range(num_epochs):

                self.total_epochs += 1
                self.normalize_errors()

                self.set_algorithm(algo_string = self.algo_string, population_size = self.population_size)

                new_variables = self.ask()
                new_errors = []

                if self.pool is None:
                    for variable in new_variables:
                        variable_dict = dict(zip(self.variable_names, variable))
                        error = self.error_function(variable_dict, self.forcefield_template)
                        new_errors.append(error)
                else:
                    if self.pool_type == 'mpi':
                        if not self.pool.is_master():
                            pool.wait()

                    variable_dict_list = [dict(zip(self.variable_names, variable)) for variable in new_variables]
                    new_errors = self.pool.map(partial(self.error_function, template = self.forcefield_template), variable_dict_list)

                for variable_id, variable in enumerate(new_variables):
                    self.variables.append(variable)
                    self.errors.append(new_errors[variable_id])

                self._write_out_forcefields()


        elif self.algo_framework == 'pymoo':
            for epoch in range(num_epochs):

                self.total_epochs += 1
                self.normalize_errors()

                self.set_algorithm(algo_string = self.algo_string, population_size = self.population_size)

                new_variables = self.ask()
                new_errors = []

                if self.pool is None:
                    for variable in new_variables:
                        variable_dict = dict(zip(self.variable_names, variable))
                        error = self.error_function(variable_dict, self.forcefield_template)
                        new_errors.append(error)
                else:
                    if self.pool_type == 'mpi':
                        if not self.pool.is_master():
                            pool.wait()

                    variable_dict_list = [dict(zip(self.variable_names, variable)) for variable in new_variables]
                    new_errors = self.pool.map(partial(self.error_function, self.forcefield_template), variable_dict_list)

                for variable_id, variable in enumerate(new_variables):
                    self.variables.append(variable)
                    self.errors.append(new_errors[variable_id])

                self._write_out_forcefields()


        elif self.algo_framework == 'platypus':
            for epoch in range(num_epochs):

                self.total_epochs += 1
                self.normalize_errors()

                self.set_algorithm(algo_string = self.algo_string, population_size = self.population_size)

                new_variables = self.ask()
                new_errors = []

                if self.pool is None:
                    for variable in new_variables:
                        variable_dict = dict(zip(self.variable_names, variable))
                        error = self.error_function(variable_dict, self.forcefield_template)
                        new_errors.append(error)

                    # for variable_id, variable in enumerate(new_variables):
                    #     self.variables.append(variable)
                    #     self.errors.append(new_errors[variable_id])

                else:
                    if self.pool_type == 'mpi':
                        if not self.pool.is_master():
                            pool.wait()

                    variable_dict_list = [dict(zip(self.variable_names, variable)) for variable in new_variables]
                    new_errors = self.pool.map(partial(self.error_function, template = self.forcefield_template), variable_dict_list)

                    # for variable_id, variable in enumerate(new_variables):
                    #     self.variables.append(variable)
                    #     self.errors.append(new_errors[variable_id])


                for variable_id, variable in enumerate(new_variables):
                    self.variables.append(variable)
                    self.errors.append(new_errors[variable_id])

                self._write_out_forcefields()



    def _write_out_forcefields(self):
        print('Epoch: '+ str(self.total_epochs))

        if not (self.algo_string.upper() == 'RANDOMSEARCH_SO'):
            # Make output files/directories
            outdir = 'results/' + str(self.total_epochs)
            if not os.path.isdir(outdir):
                os.makedirs(outdir)

            varfilename = outdir + '/variables'
            errfilename = outdir + '/errors'
            varfile = open(varfilename, 'w')
            errfile = open(errfilename, 'w')

            reco_vars, reco_errs = self.get_best_recommendation()

            if self.algo_framework == 'nevergrad' and self.num_errors == 1:
                reco_vars = [[reco_vars[key] for key in self.variable_names]]
                reco_errs = [[reco_errs]]

            for reco_id, reco_var in enumerate(reco_vars):
                varfile.write(' '.join([str(variable) for variable in reco_var]))
                varfile.write('\n')
                if self.algo_framework == 'mobopt':
                    errfile.write(' '.join([str(0.0 - error) for error in reco_errs[reco_id]]))
                else:
                    errfile.write(' '.join([str(error) for error in reco_errs[reco_id]]))
                errfile.write('\n')
            varfile.close()
            errfile.close()

            all_evaluated_filename = outdir + '/all_evaluated_'
            self.save_evaluated(all_evaluated_filename)

            if self.total_epochs % 5 == 0:
                if not os.path.isdir(outdir+'/forcefields'):
                    os.makedirs(outdir+'/forcefields')
                for reco_id, reco_var in enumerate(reco_vars):
                #for sol_index, solution in enumerate(unique(nondominated(algorithm_for_this_stage.result))):
                    ff_name = outdir + '/forcefields/FF_' + str(reco_id+1)
                    parameters_dict = dict(zip(self.variable_names, reco_var))
                    generate_forcefield(self.forcefield_template, parameters_dict, outfile=ff_name)






    def get_best_recommendation(self):
        """
        Return the best variables evaluated so far
        """
        best_variables = None
        best_errors = None
        if self.algo_framework == 'nevergrad':
            best_recommendation = self.algorithm.provide_recommendation()
            best_variables = best_recommendation.value
            best_errors = best_recommendation.loss
        elif self.algo_framework == 'mobopt':
            best_errors, best_variables = self.algorithm.space.ParetoSet()
        elif self.algo_framework == 'pymoo':
            self.normalize_errors()
            initial_population = []
            for varid, var in enumerate(self.variables):
                evaledsoln = pymoo_Individual()
                evaledsoln._X = var
                evaledsoln._F = self.normalized_errors[varid]
                initial_population.append(evaledsoln)
            initial_population = pymoo_Population(initial_population)
            if '_SO_' in self.algo_string:
                self.algorithm.pop = initial_population
                self.algorithm._set_optimum()
            else:
                self.algorithm.tell(infills = initial_population)
            best_recommendation = self.algorithm.result()
            best_variables = best_recommendation.X
            best_errors = best_recommendation.F
        elif self.algo_framework == 'platypus':
            platypus.nondominated_sort(self.algorithm.population)
            recommendation = platypus.unique(platypus.nondominated(self.algorithm.population))
            best_variables = []
            best_errors = []
            recommendation = list(set(recommendation))   # Remove duplicates
            for single_reco in recommendation:
                best_variables.append(single_reco.variables[:])
                best_errors.append(single_reco.objectives[:])

        return [best_variables, best_errors]


    def generate_pool(self, pool_type):
        """
        Return a parallel pool object

        :param pool_type: Type of parallel pool
        :type pool_type: str
        """
        if pool_type == 'multi':
            self.pool_type = 'multi'
            return sch_MultiPool()
        elif pool_type == 'mpi':
            self.pool_type = 'mpi'
            return sch_MPIPool()

    def save_evaluated(self,filename):
        """
        Save all variables evaluated so far as a zipped numpy array

        :param filename: File to which variables are saved
        :type filename: str
        """
        timestamp = datetime.now().strftime("%y%m%d_%H%M%S")
        filename = filename+timestamp+'.npz'
        np.savez(filename, variables=np.array(self.variables), errors=np.array(self.errors))

    def load_evaluated(self,filename):
        """
        Load all variables evaluated from a zipped numpy array

        :param filename: File to load variables from
        :type filename: str
        """
        fileobj = np.load(filename)
        vars = fileobj['variables']
        errs = fileobj['errors']
        for varid, var in enumerate(vars):
            self.variables.append(var)
            self.errors.append(errs[varid])


    def normalize_errors(self, scale = 2.0):
        self.normalized_errors = []
        if len(self.errors) > 0:
            maximums = np.nanmax(self.errors, axis=0)
            maximums[np.isnan(maximums)] = 100.0
            for var in self.errors:
                if np.any(np.isnan(var)):
                    self.normalized_errors.append(maximums * scale)
                else:
                    self.normalized_errors.append(var)

        self.normalized_errors = np.array(self.normalized_errors)
        if self.algo_framework == 'mobopt':
            self.normalized_errors = 0.0 - self.normalized_errors    # MOBOPT will only attempt to maximize the error function



def get_pool_rank():
    """
    Return the rank of the current process in a parallel setting
    """
    try:
        myrank = multiprocessing.current_process()._identity[0]
        return myrank                     # return rank of multiprocessing process, if available
    except:
        return MPI.COMM_WORLD.Get_rank()  # return rank of MPI process, or 0 if there is no MPI



"""Interface to LAMMPS, the Large-scale Atomic/Molecular Massively Parallel Simulator"""
import os
import xtal
import numpy as np
import ezff
from ezff.ffio import generate_forcefield as gen_ff
from ezff.utils import convert_units as convert
from ezff.utils import atomic_properties
import distutils
from distutils import spawn


class job:
    """
    Class representing a LAMMPS calculation
    """
    def __init__(self, verbose=False, path='.'):
        """
        :param path: Path where the GULP job must be run from
        :type path: str

        :param verbose: Print details about the GULP job
        :type verbose: bool
        """
        if not os.path.isdir(path):
            if verbose:
                print('Path for current job is not valid . Creating a new directory...')
            os.makedirs(path)
        self.path = path

        self.scriptfile = os.path.join(os.path.abspath(path), 'in.lmp')
        self.outfile = os.path.join(os.path.abspath(path), 'out.lmp')
        self.dumpfile = os.path.join(os.path.abspath(path), 'out.dump')
        self.structfile = os.path.join(os.path.abspath(path), 'input.structure')
        self.structure = None
        self.forcefieldfile = os.path.join(os.path.abspath(path), 'generated_forcefield')
        self.forcefield = ''
        self.options = {
            "relax_atoms": False,
            "relax_cell": False,
            "pbc": False,
            "atomic_charges": False,
            "phonon_dispersion": None,
            "phonon_dispersion_from": None,
            "phonon_dispersion_to": None
            }
        self.pbc = True
        self.verbose = verbose
        if verbose:
            print('Created a new LAMMPS job')


    def generate_forcefield(self, template_string, parameters, FFtype = None, outfile = None):
        self.options['fftype'] = FFtype.upper()
        forcefield = gen_ff(template_string, parameters, FFtype = FFtype, outfile = outfile, MD = 'LAMMPS')
        return forcefield


    def run_static(self, command = None, timeout = None):
        """
        Execute LAMMPS job with user-defined parameters

        :param command: path to LAMMPS executable
        :type command: str

        :param timeout: LAMMPS job is automatically killed after ``timeout`` seconds
        :type timeout: int
        """
        self._pre_job_cleanup()
        for snapID, snapshot in enumerate(self.structure.snaplist):
            self._write_structure_file(snapID)
            self._write_script_file_static()
            self._run_loop(command = command, timeout = timeout)
        self._post_job_cleanup()



    def run_relax(self, command = None, timeout = None):
        """
        Execute LAMMPS job with user-defined parameters

        :param command: path to LAMMPS executable
        :type command: str

        :param timeout: LAMMPS job is automatically killed after ``timeout`` seconds
        :type timeout: int
        """
        self._pre_job_cleanup()
        for snapID, snapshot in enumerate(self.structure.snaplist):
            self._write_structure_file(snapID)
            self._write_script_file_relax()
            self._run_loop(command = command, timeout = timeout)
        self._post_job_cleanup()



    def run_elastic(self, command = None, timeout = None):
        """
        Execute LAMMPS job with user-defined parameters

        :param command: path to LAMMPS executable
        :type command: str

        :param timeout: LAMMPS job is automatically killed after ``timeout`` seconds
        :type timeout: int
        """
        self._pre_job_cleanup()
        for snapID, snapshot in enumerate(self.structure.snaplist):
            self._write_structure_file(snapID)
            self._write_script_file_elastic()
            self._run_loop(command = command, timeout = timeout)
        self._post_job_cleanup()



    def _pre_job_cleanup(self):
        os.system('cd '+ self.path + ' ; rm -f ' + self.outfile)


    def _run_loop(self, command = None, timeout = None):
        outfile_single_snapshot = self.outfile + '_temp'

        system_call_command = command + ' -in ' + self.scriptfile + ' > ' + outfile_single_snapshot + ' 2> ' + self.outfile + '.runerror'

        if timeout is not None:
            system_call_command = 'timeout ' + str(timeout) + ' ' + system_call_command

        if self.verbose:
            print('cd '+ self.path + ' ; ' + system_call_command)
        os.system('cd '+ self.path + ' ; ' + system_call_command)

        # Append output to the job outfile and structure to job structurefile
        os.system('cat ' + outfile_single_snapshot + ' >> ' + self.outfile)
        os.system('cat ' + os.path.join(os.path.abspath(self.path), 'tempdumpfile') + ' >> ' + self.dumpfile)
        os.system('rm -f ' + os.path.join(os.path.abspath(self.path), 'tempdumpfile'))
        os.system('rm -f ' + outfile_single_snapshot)


    def _post_job_cleanup(self):
        files_to_be_removed = [self.outfile+'.disp', self.outfile+'.dens', self.outfile+'.runerror']
        for file_to_be_removed in files_to_be_removed:
            abs_file_to_be_removed = os.path.join(os.path.abspath(self.path), file_to_be_removed)
            os.system('cd '+ self.path + ' ; rm -f ' + abs_file_to_be_removed)




    def run(self, command = None, timeout = None):
        """
        Execute LAMMPS job with user-defined parameters

        :param command: path to LAMMPS executable
        :type command: str

        :param timeout: LAMMPS job is automatically killed after ``timeout`` seconds
        :type timeout: int
        """
        if command is None:
            # Attempt to locate a `gulp` executable
            lmpexec = distutils.spawn.find_executable('lmp_serial')
            if lmpexec is not None:
                print('Located LAMMPS executable at ' + lmpexec)
                command = lmpexec
            else:
                print('No LAMMPS executable specified or located')

        ## Purge existing result files
        os.system('cd '+ self.path + ' ; rm ' + self.outfile)

        outfile_single_snapshot = self.outfile + '_temp'

        for snapID, snapshot in enumerate(self.structure.snaplist):
            self._write_structure_file(snapID)
            self._write_script_file()

            system_call_command = command + ' -in ' + self.scriptfile + ' > ' + outfile_single_snapshot + ' 2> ' + self.outfile + '.runerror'

            if timeout is not None:
                system_call_command = 'timeout ' + str(timeout) + ' ' + system_call_command

            if self.verbose:
                print('cd '+ self.path + ' ; ' + system_call_command)
            os.system('cd '+ self.path + ' ; ' + system_call_command)

            # Append output to the job outfile and structure to job structurefile
            os.system('cat ' + outfile_single_snapshot + ' >> ' + self.outfile)
            os.system('cat ' + os.path.join(os.path.abspath(self.path), 'tempdumpfile') + ' >> ' + self.dumpfile)
            os.system('rm -f tempdumpfile')


    def _write_structure_file(self, snap_ID):
        """
        Write-out a complete structure file for the LAMMPS ``read_data`` command, based on job parameters
        """
        opts = self.options
        self.structure.box_to_abc()
        snapshot = self.structure.snaplist[snap_ID]

        structfile = open(self.structfile,'w')

        ## Define mapping between atomic elements and numeric IDs --> Only numeric IDs are allowed in structure files. Remapping is done in the script file
        map_element_id = {}
        atom_types = set([atom.element for atom in snapshot.atomlist])
        for ID, atom_type in enumerate(atom_types):
            map_element_id.update({atom_type: ID+1})

        self.options['atom_sequence'] = list(atom_types)

        # Write header for the data file
        structfile.write('## LAMMPS structure file from EZFF \n\n')
        structfile.write(str(len(snapshot.atomlist)) + ' atoms \n\n')
        structfile.write(str(len(atom_types)) + ' atom types \n\n')


        # Coordinate transformation (taken from https://lammps.sandia.gov/doc/Howto_triclinic.html)
        self.structure.box_to_abc()
        abc = self.structure.abc
        ang = self.structure.ang
        lx = abc[0]
        xy = abc[1] * np.cos(ang[2])
        xz = abc[2] * np.cos(ang[1])
        ly = np.sqrt((abc[1]*abc[1]) - (xy*xy))
        yz = ((abc[1]*(abc[2]*np.cos(ang[0]))) - (xy*xz))/ly
        lz = np.sqrt((abc[2]*abc[2]) - (xz*xz) - (yz*yz))


        # Correct for very large skew of the box
        xy = np.minimum(np.maximum(xy, 0.0-lx/2.0), lx/2.0)
        xz = np.minimum(np.maximum(xz, 0.0-lx/2.0), lx/2.0)
        yz = np.minimum(np.maximum(yz, 0.0-ly/2.0), ly/2.0)

        xy = np.round(xy, decimals=4)
        xz = np.round(xz, decimals=4)
        yz = np.round(yz, decimals=4)

        # if ((abs(xy) < 1e-4) and (abs(yz) < 1e-4) and (abs(xz) < 1e-4)):
        #     structfile.write('0.0 ' + str(lx) + ' xlo xhi \n')
        #     structfile.write('0.0 ' + str(ly) + ' ylo yhi \n')
        #     structfile.write('0.0 ' + str(lz) + ' zlo zhi \n')
        # else:
        #     structfile.write('0.0 ' + str(lx) + ' xlo xhi \n')
        #     structfile.write('0.0 ' + str(ly) + ' ylo yhi \n')
        #     structfile.write('0.0 ' + str(lz) + ' zlo zhi \n')
        #     structfile.write(str(xy) + ' ' + str(xz) + ' ' + str(yz) + ' xy xz yz \n')


        structfile.write('0.0 ' + str(lx) + ' xlo xhi \n')
        structfile.write('0.0 ' + str(ly) + ' ylo yhi \n')
        structfile.write('0.0 ' + str(lz) + ' zlo zhi \n')
        structfile.write(str(xy) + ' ' + str(xz) + ' ' + str(yz) + ' xy xz yz \n')

        structfile.write('\n\n')

        structfile.write('Masses\n\n')
        for ID, atom_type in enumerate(atom_types):
            structfile.write('%d  %f \n' % (ID+1, atomic_properties.atomic_mass[atom_type]))
        structfile.write('\n\n')

        structfile.write('Atoms\n\n')
        for atomID, atom in enumerate(snapshot.atomlist):
            structfile.write('%d %d %f %f %f %f \n' % (atomID+1, map_element_id[atom.element], atom.charge, atom.cart[0], atom.cart[1], atom.cart[2]))


    def _write_forcefield_file(self):
        fffile = open(os.path.join(os.path.abspath(self.path), 'ff.lmp'),'w')
        fffile.write(self.forcefield)
        fffile.close()


        script.write('read_data input.structure \n')
        script.write('pair_style ' + fftype_pairstyle[opts['fftype']] + '\n')
        script.write('pair_coeff * * ff.lmp ' + ' '.join(opts['atom_sequence']).title() + '\n')

        ## Define a universal thermo style and dump information every step
        script.write('thermo_style custom step temp pxx pyy pzz pxy pxz pyz pe ke etotal vol xlo xhi ylo yhi zlo zhi xy xz yz press lx ly lz \n')
        script.write('thermo 1000 \n')

        script.write('variable ezff_T equal temp \n')
        script.write('variable ezff_V equal vol \n')
        script.write('variable ezff_E equal etotal \n')

        script.write('run 0 \n')

        script.write('\n')
        script.write('\n')



    def _include_forcefield(self, fftype):
        # Convert FFtype to pair_style
        fftype_pairstyle = {'SW': 'sw', 'STILLINGER-WEBER': 'sw', 'STILLINGER WEBER': 'sw',
                            'REAX': 'reax/c NULL', 'REAXFF': 'reax/c NULL', 'REAX/C': 'reax/c NULL',
                            'VASHISHTA': 'vashishta'}
        return_string = ''
        if self.options['fftype'].upper() == 'LJ':
            return_string = 'include generated_forcefield \n'
            f = open(self.forcefieldfile, 'w')
            f.write(self.forcefield)
            f.close()
        elif self.options['fftype'].upper() == 'VASHISHTA':
            return_string = 'pair_style vashishta \n'
            return_string +=  'pair_coeff * * ' + self.forcefieldfile  + ' ' + ' '.join(self.options['atom_sequence']).title() + '\n'
            f = open(self.forcefieldfile, 'w')
            f.write(self.forcefield)
            f.close()
        elif 'REAX' in self.options['fftype'].upper():
            return_string = 'pair_style ' + fftype_pairstyle[self.options['fftype']] + '\n'
            return_string += 'pair_coeff * * ' + self.forcefieldfile + ' ' + ' '.join(self.options['atom_sequence']).title() + '\n'
            return_string += 'fix 1 all qeq/reaxff 1 0.0 10.0 1.0e-6 reaxff \n' # Set charge optimization
            f = open(self.forcefieldfile, 'w')
            f.write(self.forcefield)
            f.close()
        else:
            return_string = 'pair_style ' + fftype_pairstyle[self.options['fftype']] + '\n'
            return_string += 'pair_coeff * * ' + self.forcefieldfile + ' ' + ' '.join(self.options['atom_sequence']).title() + '\n'
            f = open(self.forcefieldfile, 'w')
            f.write(self.forcefield)
            f.close()
        return return_string




    def _write_script_file_static(self):
        opts = self.options
        script = open(self.scriptfile,'w')

        script.write('units metal \n')  # All simulations will be performed in LAMMPS metal units
        script.write('dimension 3 \n')  # All simulations will be 3D
        script.write('atom_style charge \n')

        if opts['pbc']:
            script.write('boundary p p p \n')
        else:
            script.write('boundary fm fm fm \n')

        script.write('read_data input.structure \n')

        ff_include_string = self._include_forcefield(self.options['fftype'])
        script.write(ff_include_string)

        ## Define a universal thermo style and dump information every step
        script.write('thermo_style custom step temp pxx pyy pzz pxy pxz pyz pe ke etotal vol xlo xhi ylo yhi zlo zhi xy xz yz press lx ly lz \n')
        script.write('thermo 1000 \n')

        script.write('variable ezff_T equal temp \n')
        script.write('variable ezff_V equal vol \n')
        script.write('variable ezff_E equal etotal \n')

        script.write('run 0 \n')

        script.write('\n')
        script.write('\n')

        # Write out the summary
        script.write('write_dump all custom tempdumpfile id type element mass q x y z vx vy vz fx fy fz modify sort id element ' + ' '.join(opts['atom_sequence']).title() + ' \n')

        script.write('print "-----SUMMARY-----" \n')
        script.write('print "EZFF_TEMP ${ezff_T}" \n')
        script.write('print "EZFF_VOL ${ezff_V}" \n')
        script.write('print "EZFF_ENERGY ${ezff_E}" \n')

        script.close()





    def _write_script_file_relax(self):
        opts = self.options
        script = open(self.scriptfile,'w')

        script.write('units metal \n')  # All simulations will be performed in LAMMPS metal units
        script.write('dimension 3 \n')  # All simulations will be 3D
        script.write('atom_style charge \n')

        if opts['pbc']:
            script.write('boundary p p p \n')
        else:
            script.write('boundary fm fm fm \n')

        script.write('read_data input.structure \n')

        ff_include_string = self._include_forcefield(self.options['fftype'])
        script.write(ff_include_string)

        ## Define a universal thermo style and dump information every step
        script.write('thermo_style custom step temp pxx pyy pzz pxy pxz pyz pe ke etotal vol xlo xhi ylo yhi zlo zhi xy xz yz press lx ly lz \n')
        script.write('thermo 1000 \n')

        script.write('variable ezff_T equal temp \n')
        script.write('variable ezff_V equal vol \n')
        script.write('variable ezff_E equal etotal \n')

        if opts['relax_cell']:
            script.write('fix FixBoxRelax all box/relax aniso 0.0 \n')
        script.write('minimize 0.0 1.0e-8 1000 100000 \n')

        script.write('run 0 \n')

        script.write('\n')
        script.write('\n')

        # Write out the summary
        script.write('write_dump all custom tempdumpfile id type element mass q x y z vx vy vz fx fy fz modify sort id element ' + ' '.join(opts['atom_sequence']).title() + ' \n')

        script.write('print "-----SUMMARY-----" \n')
        script.write('print "EZFF_TEMP ${ezff_T}" \n')
        script.write('print "EZFF_VOL ${ezff_V}" \n')
        script.write('print "EZFF_ENERGY ${ezff_E}" \n')

        script.close()





    def _write_script_file_elastic(self):
        opts = self.options
        script = open(self.scriptfile,'w')

        script.write('units metal \n')  # All simulations will be performed in LAMMPS metal units
        script.write('dimension 3 \n')  # All simulations will be 3D
        script.write('atom_style charge \n')

        if opts['pbc']:
            script.write('boundary p p p \n')
        else:
            script.write('boundary fm fm fm \n')

        script.write('read_data input.structure \n')

        ff_include_string = self._include_forcefield(self.options['fftype'])
        script.write(ff_include_string)

        ## Define a universal thermo style and dump information every step
        script.write('thermo_style custom step temp pxx pyy pzz pxy pxz pyz pe ke etotal vol xlo xhi ylo yhi zlo zhi xy xz yz press lx ly lz \n')
        script.write('thermo 1000 \n')

        script.write('variable ezff_T equal temp \n')
        script.write('variable ezff_V equal vol \n')
        script.write('variable ezff_E equal etotal \n')

        if opts['relax_cell']:
            script.write('fix FixBoxRelax all box/relax aniso 0.0 \n')
        script.write('minimize 0.0 1.0e-8 1000 100000 \n')

        script.write('run 0 \n')

        script.write('\n')
        script.write('\n')

        # Write out the summary
        script.write('write_dump all custom tempdumpfile id type element mass q x y z vx vy vz fx fy fz modify sort id element ' + ' '.join(opts['atom_sequence']).title() + ' \n')

        script.write('print "-----SUMMARY-----" \n')
        script.write('print "EZFF_TEMP ${ezff_T}" \n')
        script.write('print "EZFF_VOL ${ezff_V}" \n')
        script.write('print "EZFF_ENERGY ${ezff_E}" \n')


        script.close()

        self._write_elastic_displacement_mod()




    def _write_elastic_displacement_mod(self):
        """
        Write-out a complete LAMMPS script file, ``job.scriptfile``, based on job parameters
        """
        opts = self.options
        script = open(self.scriptfile,'a+')

        if opts['pbc']:
            displace_script = open(os.path.join(os.path.abspath(self.path), 'displace.mod'), 'w')

            displace_script.write('clear\nvariable up equal 1.0e-6\nvariable atomjiggle equal 1.0e-5\nvariable cfac equal 1.0e-4\nvariable cunits string GPa\nif "${dir} == 1" then &\n   "variable len0 equal ${lx0}"\nif "${dir} == 2" then &\n   "variable len0 equal ${ly0}"\nif "${dir} == 3" then &\n   "variable len0 equal ${lz0}"\nif "${dir} == 4" then &\n   "variable len0 equal ${lz0}"\nif "${dir} == 5" then &\n   "variable len0 equal ${lz0}"\nif "${dir} == 6" then &\n   "variable len0 equal ${ly0}"\nbox tilt large\nread_restart restart.equil\n')

            ff_include_string = self._include_forcefield(self.options['fftype'])
            displace_script.write(ff_include_string)
            # displace_script.write('pair_style ' + fftype_pairstyle[opts['fftype']] + '\n')
            # displace_script.write('pair_coeff * * ff.lmp ' + ' '.join(opts['atom_sequence']).title() + '\n')
            displace_script.write('thermo_style custom step temp pxx pyy pzz pxy pxz pyz pe ke etotal vol xlo xhi ylo yhi zlo zhi xy xz yz press lx ly lz \n')
            displace_script.write('thermo 1 \n')

            displace_script.write('variable delta equal -${up}*${len0}\nvariable deltaxy equal -${up}*xy\nvariable deltaxz equal -${up}*xz\nvariable deltayz equal -${up}*yz\nif "${dir} == 1" then &\n   "change_box all x delta 0 ${delta} xy delta ${deltaxy} xz delta ${deltaxz} remap units box"\nif "${dir} == 2" then &\n   "change_box all y delta 0 ${delta} yz delta ${deltayz} remap units box"\nif "${dir} == 3" then &\n   "change_box all z delta 0 ${delta} remap units box"\nif "${dir} == 4" then &\n   "change_box all yz delta ${delta} remap units box"\nif "${dir} == 5" then &\n   "change_box all xz delta ${delta} remap units box"\nif "${dir} == 6" then &\n   "change_box all xy delta ${delta} remap units box"\nminimize 0.0 1.0e-8 1000 100000\nvariable tmp equal pxx\nvariable pxx1 equal ${tmp}\nvariable tmp equal pyy\nvariable pyy1 equal ${tmp}\nvariable tmp equal pzz\nvariable pzz1 equal ${tmp}\nvariable tmp equal pxy\nvariable pxy1 equal ${tmp}\nvariable tmp equal pxz\nvariable pxz1 equal ${tmp}\nvariable tmp equal pyz\nvariable pyz1 equal ${tmp}\nvariable C1neg equal ${d1}\nvariable C2neg equal ${d2}\nvariable C3neg equal ${d3}\nvariable C4neg equal ${d4}\nvariable C5neg equal ${d5}\nvariable C6neg equal ${d6}\nclear\nbox tilt large\nread_restart restart.equil\n')

            ff_include_string = self._include_forcefield(self.options['fftype'])
            displace_script.write(ff_include_string)
            # displace_script.write('pair_style ' + fftype_pairstyle[opts['fftype']] + '\n')
            # displace_script.write('pair_coeff * * ff.lmp ' + ' '.join(opts['atom_sequence']).title() + '\n')
            displace_script.write('thermo_style custom step temp pxx pyy pzz pxy pxz pyz pe ke etotal vol xlo xhi ylo yhi zlo zhi xy xz yz press lx ly lz \n')
            displace_script.write('thermo 1 \n')

            displace_script.write('variable delta equal ${up}*${len0}\nvariable deltaxy equal ${up}*xy\nvariable deltaxz equal ${up}*xz\nvariable deltayz equal ${up}*yz\nif "${dir} == 1" then &\n   "change_box all x delta 0 ${delta} xy delta ${deltaxy} xz delta ${deltaxz} remap units box"\nif "${dir} == 2" then &\n   "change_box all y delta 0 ${delta} yz delta ${deltayz} remap units box"\nif "${dir} == 3" then &\n   "change_box all z delta 0 ${delta} remap units box"\nif "${dir} == 4" then &\n   "change_box all yz delta ${delta} remap units box"\nif "${dir} == 5" then &\n   "change_box all xz delta ${delta} remap units box"\nif "${dir} == 6" then &\n   "change_box all xy delta ${delta} remap units box"\nminimize 0.0 1.0e-8 1000 100000\nvariable tmp equal pe\nvariable e1 equal ${tmp}\nvariable tmp equal press\nvariable p1 equal ${tmp}\nvariable tmp equal pxx\nvariable pxx1 equal ${tmp}\nvariable tmp equal pyy\nvariable pyy1 equal ${tmp}\nvariable tmp equal pzz\nvariable pzz1 equal ${tmp}\nvariable tmp equal pxy\nvariable pxy1 equal ${tmp}\nvariable tmp equal pxz\nvariable pxz1 equal ${tmp}\nvariable tmp equal pyz\nvariable pyz1 equal ${tmp}\nvariable C1pos equal ${d1}\nvariable C2pos equal ${d2}\nvariable C3pos equal ${d3}\nvariable C4pos equal ${d4}\nvariable C5pos equal ${d5}\nvariable C6pos equal ${d6}\nvariable C1${dir} equal 0.5*(${C1neg}+${C1pos})\nvariable C2${dir} equal 0.5*(${C2neg}+${C2pos})\nvariable C3${dir} equal 0.5*(${C3neg}+${C3pos})\nvariable C4${dir} equal 0.5*(${C4neg}+${C4pos})\nvariable C5${dir} equal 0.5*(${C5neg}+${C5pos})\nvariable C6${dir} equal 0.5*(${C6neg}+${C6pos})\nvariable dir delete\n')

            displace_script.close()

            script.write('\nvariable up equal 1.0e-6\nvariable atomjiggle equal 1.0e-5\nvariable cfac equal 1.0e-4\nvariable cunits string GPa\nvariable tmp equal pxx\nvariable pxx0 equal ${tmp}\nvariable tmp equal pyy\nvariable pyy0 equal ${tmp}\nvariable tmp equal pzz\nvariable pzz0 equal ${tmp}\nvariable tmp equal pyz\nvariable pyz0 equal ${tmp}\nvariable tmp equal pxz\nvariable pxz0 equal ${tmp}\nvariable tmp equal pxy\nvariable pxy0 equal ${tmp}\nvariable tmp equal lx\nvariable lx0 equal ${tmp}\nvariable tmp equal ly\nvariable ly0 equal ${tmp}\nvariable tmp equal lz\nvariable lz0 equal ${tmp}\n# These formulas define the derivatives w.r.t. strain components\n# Constants uses $, variables use v_\nvariable d1 equal -(v_pxx1-${pxx0})/(v_delta/v_len0)*${cfac}\nvariable d2 equal -(v_pyy1-${pyy0})/(v_delta/v_len0)*${cfac}\nvariable d3 equal -(v_pzz1-${pzz0})/(v_delta/v_len0)*${cfac}\nvariable d4 equal -(v_pyz1-${pyz0})/(v_delta/v_len0)*${cfac}\nvariable d5 equal -(v_pxz1-${pxz0})/(v_delta/v_len0)*${cfac}\nvariable d6 equal -(v_pxy1-${pxy0})/(v_delta/v_len0)*${cfac}\n')

            if opts['relax_cell']:
                script.write('unfix FixBoxRelax\n')
            script.write('write_restart restart.equil\n')

            script.write('variable dir equal 1\ninclude displace.mod\nvariable dir equal 2\ninclude displace.mod\nvariable dir equal 3\ninclude displace.mod\nvariable dir equal 4\ninclude displace.mod\nvariable dir equal 5\ninclude displace.mod\nvariable dir equal 6\ninclude displace.mod\nvariable C11all equal ${C11}\nvariable C22all equal ${C22}\nvariable C33all equal ${C33}\nvariable C12all equal 0.5*(${C12}+${C21})\nvariable C13all equal 0.5*(${C13}+${C31})\nvariable C23all equal 0.5*(${C23}+${C32})\nvariable C44all equal ${C44}\nvariable C55all equal ${C55}\nvariable C66all equal ${C66}\nvariable C14all equal 0.5*(${C14}+${C41})\nvariable C15all equal 0.5*(${C15}+${C51})\nvariable C16all equal 0.5*(${C16}+${C61})\nvariable C24all equal 0.5*(${C24}+${C42})\nvariable C25all equal 0.5*(${C25}+${C52})\nvariable C26all equal 0.5*(${C26}+${C62})\nvariable C34all equal 0.5*(${C34}+${C43})\nvariable C35all equal 0.5*(${C35}+${C53})\nvariable C36all equal 0.5*(${C36}+${C63})\nvariable C45all equal 0.5*(${C45}+${C54})\nvariable C46all equal 0.5*(${C46}+${C64})\nvariable C56all equal 0.5*(${C56}+${C65})\nvariable C11cubic equal (${C11all}+${C22all}+${C33all})/3.0\nvariable C12cubic equal (${C12all}+${C13all}+${C23all})/3.0\nvariable C44cubic equal (${C44all}+${C55all}+${C66all})/3.0\nvariable bulkmodulus equal (${C11cubic}+2*${C12cubic})/3.0\nvariable shearmodulus1 equal ${C44cubic}\nvariable shearmodulus2 equal (${C11cubic}-${C12cubic})/2.0\nvariable poissonratio equal 1.0/(1.0+${C11cubic}/${C12cubic})\nprint "EZFF C11 ${C11all} ${cunits}"\nprint "EZFF C22 ${C22all} ${cunits}"\nprint "EZFF C33 ${C33all} ${cunits}"\nprint "EZFF C12 ${C12all} ${cunits}"\nprint "EZFF C13 ${C13all} ${cunits}"\nprint "EZFF C23 ${C23all} ${cunits}"\nprint "EZFF C44 ${C44all} ${cunits}"\nprint "EZFF C55 ${C55all} ${cunits}"\nprint "EZFF C66 ${C66all} ${cunits}"\nprint "EZFF C14 ${C14all} ${cunits}"\nprint "EZFF C15 ${C15all} ${cunits}"\nprint "EZFF C16 ${C16all} ${cunits}"\nprint "EZFF C24 ${C24all} ${cunits}"\nprint "EZFF C25 ${C25all} ${cunits}"\nprint "EZFF C26 ${C26all} ${cunits}"\nprint "EZFF C34 ${C34all} ${cunits}"\nprint "EZFF C35 ${C35all} ${cunits}"\nprint "EZFF C36 ${C36all} ${cunits}"\nprint "EZFF C45 ${C45all} ${cunits}"\nprint "EZFF C46 ${C46all} ${cunits}"\nprint "EZFF C56 ${C56all} ${cunits}"\nprint "EZFF Bulk_Modulus ${bulkmodulus} ${cunits}"\nprint "EZFF Shear_Modulus_1 ${shearmodulus1} ${cunits}"\nprint "EZFF Shear_Modulus_2 ${shearmodulus2} ${cunits}"\nprint "EZFF Poisson_Ratio ${poissonratio}"\n')


        script.close()






    def _write_script_file(self):
        """
        Write-out a complete LAMMPS script file, ``job.scriptfile``, based on job parameters
        """
        opts = self.options
        script = open(self.scriptfile,'w')

        script.write('units metal \n')  # All simulations will be performed in LAMMPS metal units
        script.write('dimension 3 \n')  # All simulations will be 3D
        script.write('atom_style charge \n')

        if opts['pbc']:
            script.write('boundary p p p \n')
        else:
            script.write('boundary fm fm fm \n')

        script.write('read_data input.structure \n')

        ff_include_string = self._include_forcefield(self.options['fftype'])
        script.write(ff_include_string)
        # script.write('pair_style ' + fftype_pairstyle[opts['fftype']] + '\n')
        # script.write('pair_coeff * * ff.lmp ' + ' '.join(opts['atom_sequence']).title() + '\n')

        ## Define a universal thermo style and dump information every step
        script.write('thermo_style custom step temp pxx pyy pzz pxy pxz pyz pe ke etotal vol xlo xhi ylo yhi zlo zhi xy xz yz press lx ly lz \n')
        script.write('thermo 1000 \n')

        script.write('variable ezff_T equal temp \n')
        script.write('variable ezff_V equal vol \n')
        script.write('variable ezff_E equal etotal \n')

        if opts['relax_atoms']:
            if opts['relax_cell']:
                script.write('fix FixBoxRelax all box/relax aniso 0.0 \n')
            script.write('minimize 0.0 1.0e-8 1000 100000 \n')
        else:
            script.write('run 0 \n')

        script.write('\n')
        script.write('\n')

        # Write out the summary
        script.write('write_dump all custom tempdumpfile id type element mass q x y z vx vy vz fx fy fz modify sort id element ' + ' '.join(opts['atom_sequence']).title() + ' \n')

        script.write('print "-----SUMMARY-----" \n')
        script.write('print "EZFF_TEMP ${ezff_T}" \n')
        script.write('print "EZFF_VOL ${ezff_V}" \n')
        script.write('print "EZFF_ENERGY ${ezff_E}" \n')

        # SCRIPTS TO COMPUTE ELASTIC CONSTANT
        # ADOPTED FROM THE ELASTIC EXAMPLE IN LAMMPS DIRECTORY
        #if opts['pbc']:
        if 1==2:
            displace_script = open(os.path.join(os.path.abspath(self.path), 'displace.mod'), 'w')

            displace_script.write('clear\nvariable up equal 1.0e-6\nvariable atomjiggle equal 1.0e-5\nvariable cfac equal 1.0e-4\nvariable cunits string GPa\nif "${dir} == 1" then &\n   "variable len0 equal ${lx0}"\nif "${dir} == 2" then &\n   "variable len0 equal ${ly0}"\nif "${dir} == 3" then &\n   "variable len0 equal ${lz0}"\nif "${dir} == 4" then &\n   "variable len0 equal ${lz0}"\nif "${dir} == 5" then &\n   "variable len0 equal ${lz0}"\nif "${dir} == 6" then &\n   "variable len0 equal ${ly0}"\nbox tilt large\nread_restart restart.equil\n')

            displace_script.write('pair_style ' + fftype_pairstyle[opts['fftype']] + '\n')
            displace_script.write('pair_coeff * * ff.lmp ' + ' '.join(opts['atom_sequence']).title() + '\n')
            displace_script.write('thermo_style custom step temp pxx pyy pzz pxy pxz pyz pe ke etotal vol xlo xhi ylo yhi zlo zhi xy xz yz press lx ly lz \n')
            displace_script.write('thermo 1 \n')

            displace_script.write('variable delta equal -${up}*${len0}\nvariable deltaxy equal -${up}*xy\nvariable deltaxz equal -${up}*xz\nvariable deltayz equal -${up}*yz\nif "${dir} == 1" then &\n   "change_box all x delta 0 ${delta} xy delta ${deltaxy} xz delta ${deltaxz} remap units box"\nif "${dir} == 2" then &\n   "change_box all y delta 0 ${delta} yz delta ${deltayz} remap units box"\nif "${dir} == 3" then &\n   "change_box all z delta 0 ${delta} remap units box"\nif "${dir} == 4" then &\n   "change_box all yz delta ${delta} remap units box"\nif "${dir} == 5" then &\n   "change_box all xz delta ${delta} remap units box"\nif "${dir} == 6" then &\n   "change_box all xy delta ${delta} remap units box"\nminimize 0.0 1.0e-8 1000 100000\nvariable tmp equal pxx\nvariable pxx1 equal ${tmp}\nvariable tmp equal pyy\nvariable pyy1 equal ${tmp}\nvariable tmp equal pzz\nvariable pzz1 equal ${tmp}\nvariable tmp equal pxy\nvariable pxy1 equal ${tmp}\nvariable tmp equal pxz\nvariable pxz1 equal ${tmp}\nvariable tmp equal pyz\nvariable pyz1 equal ${tmp}\nvariable C1neg equal ${d1}\nvariable C2neg equal ${d2}\nvariable C3neg equal ${d3}\nvariable C4neg equal ${d4}\nvariable C5neg equal ${d5}\nvariable C6neg equal ${d6}\nclear\nbox tilt large\nread_restart restart.equil\n')

            displace_script.write('pair_style ' + fftype_pairstyle[opts['fftype']] + '\n')
            displace_script.write('pair_coeff * * ff.lmp ' + ' '.join(opts['atom_sequence']).title() + '\n')
            displace_script.write('thermo_style custom step temp pxx pyy pzz pxy pxz pyz pe ke etotal vol xlo xhi ylo yhi zlo zhi xy xz yz press lx ly lz \n')
            displace_script.write('thermo 1 \n')

            displace_script.write('variable delta equal ${up}*${len0}\nvariable deltaxy equal ${up}*xy\nvariable deltaxz equal ${up}*xz\nvariable deltayz equal ${up}*yz\nif "${dir} == 1" then &\n   "change_box all x delta 0 ${delta} xy delta ${deltaxy} xz delta ${deltaxz} remap units box"\nif "${dir} == 2" then &\n   "change_box all y delta 0 ${delta} yz delta ${deltayz} remap units box"\nif "${dir} == 3" then &\n   "change_box all z delta 0 ${delta} remap units box"\nif "${dir} == 4" then &\n   "change_box all yz delta ${delta} remap units box"\nif "${dir} == 5" then &\n   "change_box all xz delta ${delta} remap units box"\nif "${dir} == 6" then &\n   "change_box all xy delta ${delta} remap units box"\nminimize 0.0 1.0e-8 1000 100000\nvariable tmp equal pe\nvariable e1 equal ${tmp}\nvariable tmp equal press\nvariable p1 equal ${tmp}\nvariable tmp equal pxx\nvariable pxx1 equal ${tmp}\nvariable tmp equal pyy\nvariable pyy1 equal ${tmp}\nvariable tmp equal pzz\nvariable pzz1 equal ${tmp}\nvariable tmp equal pxy\nvariable pxy1 equal ${tmp}\nvariable tmp equal pxz\nvariable pxz1 equal ${tmp}\nvariable tmp equal pyz\nvariable pyz1 equal ${tmp}\nvariable C1pos equal ${d1}\nvariable C2pos equal ${d2}\nvariable C3pos equal ${d3}\nvariable C4pos equal ${d4}\nvariable C5pos equal ${d5}\nvariable C6pos equal ${d6}\nvariable C1${dir} equal 0.5*(${C1neg}+${C1pos})\nvariable C2${dir} equal 0.5*(${C2neg}+${C2pos})\nvariable C3${dir} equal 0.5*(${C3neg}+${C3pos})\nvariable C4${dir} equal 0.5*(${C4neg}+${C4pos})\nvariable C5${dir} equal 0.5*(${C5neg}+${C5pos})\nvariable C6${dir} equal 0.5*(${C6neg}+${C6pos})\nvariable dir delete\n')

            displace_script.close()

            script.write('\nvariable up equal 1.0e-6\nvariable atomjiggle equal 1.0e-5\nvariable cfac equal 1.0e-4\nvariable cunits string GPa\nvariable tmp equal pxx\nvariable pxx0 equal ${tmp}\nvariable tmp equal pyy\nvariable pyy0 equal ${tmp}\nvariable tmp equal pzz\nvariable pzz0 equal ${tmp}\nvariable tmp equal pyz\nvariable pyz0 equal ${tmp}\nvariable tmp equal pxz\nvariable pxz0 equal ${tmp}\nvariable tmp equal pxy\nvariable pxy0 equal ${tmp}\nvariable tmp equal lx\nvariable lx0 equal ${tmp}\nvariable tmp equal ly\nvariable ly0 equal ${tmp}\nvariable tmp equal lz\nvariable lz0 equal ${tmp}\n# These formulas define the derivatives w.r.t. strain components\n# Constants uses $, variables use v_\nvariable d1 equal -(v_pxx1-${pxx0})/(v_delta/v_len0)*${cfac}\nvariable d2 equal -(v_pyy1-${pyy0})/(v_delta/v_len0)*${cfac}\nvariable d3 equal -(v_pzz1-${pzz0})/(v_delta/v_len0)*${cfac}\nvariable d4 equal -(v_pyz1-${pyz0})/(v_delta/v_len0)*${cfac}\nvariable d5 equal -(v_pxz1-${pxz0})/(v_delta/v_len0)*${cfac}\nvariable d6 equal -(v_pxy1-${pxy0})/(v_delta/v_len0)*${cfac}\n')

            if opts['relax_cell']:
                script.write('unfix FixBoxRelax\n')
            script.write('write_restart restart.equil\n')

            script.write('variable dir equal 1\ninclude displace.mod\nvariable dir equal 2\ninclude displace.mod\nvariable dir equal 3\ninclude displace.mod\nvariable dir equal 4\ninclude displace.mod\nvariable dir equal 5\ninclude displace.mod\nvariable dir equal 6\ninclude displace.mod\nvariable C11all equal ${C11}\nvariable C22all equal ${C22}\nvariable C33all equal ${C33}\nvariable C12all equal 0.5*(${C12}+${C21})\nvariable C13all equal 0.5*(${C13}+${C31})\nvariable C23all equal 0.5*(${C23}+${C32})\nvariable C44all equal ${C44}\nvariable C55all equal ${C55}\nvariable C66all equal ${C66}\nvariable C14all equal 0.5*(${C14}+${C41})\nvariable C15all equal 0.5*(${C15}+${C51})\nvariable C16all equal 0.5*(${C16}+${C61})\nvariable C24all equal 0.5*(${C24}+${C42})\nvariable C25all equal 0.5*(${C25}+${C52})\nvariable C26all equal 0.5*(${C26}+${C62})\nvariable C34all equal 0.5*(${C34}+${C43})\nvariable C35all equal 0.5*(${C35}+${C53})\nvariable C36all equal 0.5*(${C36}+${C63})\nvariable C45all equal 0.5*(${C45}+${C54})\nvariable C46all equal 0.5*(${C46}+${C64})\nvariable C56all equal 0.5*(${C56}+${C65})\nvariable C11cubic equal (${C11all}+${C22all}+${C33all})/3.0\nvariable C12cubic equal (${C12all}+${C13all}+${C23all})/3.0\nvariable C44cubic equal (${C44all}+${C55all}+${C66all})/3.0\nvariable bulkmodulus equal (${C11cubic}+2*${C12cubic})/3.0\nvariable shearmodulus1 equal ${C44cubic}\nvariable shearmodulus2 equal (${C11cubic}-${C12cubic})/2.0\nvariable poissonratio equal 1.0/(1.0+${C11cubic}/${C12cubic})\nprint "EZFF C11 ${C11all} ${cunits}"\nprint "EZFF C22 ${C22all} ${cunits}"\nprint "EZFF C33 ${C33all} ${cunits}"\nprint "EZFF C12 ${C12all} ${cunits}"\nprint "EZFF C13 ${C13all} ${cunits}"\nprint "EZFF C23 ${C23all} ${cunits}"\nprint "EZFF C44 ${C44all} ${cunits}"\nprint "EZFF C55 ${C55all} ${cunits}"\nprint "EZFF C66 ${C66all} ${cunits}"\nprint "EZFF C14 ${C14all} ${cunits}"\nprint "EZFF C15 ${C15all} ${cunits}"\nprint "EZFF C16 ${C16all} ${cunits}"\nprint "EZFF C24 ${C24all} ${cunits}"\nprint "EZFF C25 ${C25all} ${cunits}"\nprint "EZFF C26 ${C26all} ${cunits}"\nprint "EZFF C34 ${C34all} ${cunits}"\nprint "EZFF C35 ${C35all} ${cunits}"\nprint "EZFF C36 ${C36all} ${cunits}"\nprint "EZFF C45 ${C45all} ${cunits}"\nprint "EZFF C46 ${C46all} ${cunits}"\nprint "EZFF C56 ${C56all} ${cunits}"\nprint "EZFF Bulk_Modulus ${bulkmodulus} ${cunits}"\nprint "EZFF Shear_Modulus_1 ${shearmodulus1} ${cunits}"\nprint "EZFF Shear_Modulus_2 ${shearmodulus2} ${cunits}"\nprint "EZFF Poisson_Ratio ${poissonratio}"\n')


        script.close()



    def cleanup(self):
        """
        Clean-up after the completion of a GULP job. Deletes input, output and forcefields files
        """
        files_to_be_removed = [self.outfile+'.disp', self.outfile+'.dens', self.outfile, self.scriptfile, self.outfile+'.runerror', self.structfile, self.forcefieldfile, self.dumpfile]
        for file in files_to_be_removed:
            if os.path.isfile(file):
                os.remove(file)
            elif os.path.isfile(self.path+'/'+file):
                os.remove(self.path+'/'+file)



    ## OOP methods for reading output from GULP
    def read_energy(self):
        """
        Read energy from completed GULP job

        :returns: Energy of the input structure(s) in eV as a np.ndarray
        """
        return _read_energy(self.outfile)

    def read_elastic_moduli(self):
        """
        Read elastic modulus matrix from a completed GULP job

        :returns: 6x6 Elastic modulus matrix in GPa for each input structure, as a list
        """
        return _read_elastic_moduli(self.outfile)

    def read_atomic_charges(self):
        """
        Read atomic charge information from a completed GULP job file

        :returns: xtal.AtTraj object with optimized charge information
        """
        return _read_structure(self.dumpfile) # read_structure already returns xtal object with charges

    def read_structure(self):
        """
        Read converged structure (cell and atomic positions) from the MD job

        :returns: xtal.AtTraj object with (optimized) individual structures as separate snapshots
        """
        return _read_structure(self.dumpfile)



def _read_elastic_moduli(outfilename):
    """
    Read elastic modulus matrix from a completed GULP job

    :param outfilename: Path of the stdout from the GULP job
    :type outfilename: str
    :returns: 6x6 Elastic modulus matrix in GPa
    """
    outfile = open(outfilename,'r')
    moduli_array = []
    moduli = np.zeros((6,6))
    while True:
        oneline = outfile.readline()
        if not oneline: # break at EOF
            break
        if 'EZFF C11' in oneline:
            moduli[0,0] = float(oneline.strip().split()[2]) # in Bars
            newline = outfile.readline() # C22
            moduli[1,1] = float(newline.strip().split()[2]) # in Bars
            newline = outfile.readline() # C33
            moduli[2,2] = float(newline.strip().split()[2]) # in Bars

            newline = outfile.readline() # C12
            moduli[0,1] = float(newline.strip().split()[2]) # in Bars
            moduli[1,0] = float(newline.strip().split()[2]) # in Bars
            newline = outfile.readline() # C13
            moduli[0,2] = float(newline.strip().split()[2]) # in Bars
            moduli[2,0] = float(newline.strip().split()[2]) # in Bars
            newline = outfile.readline() # C23
            moduli[1,2] = float(newline.strip().split()[2]) # in Bars
            moduli[2,1] = float(newline.strip().split()[2]) # in Bars

            newline = outfile.readline() # C44
            moduli[3,3] = float(newline.strip().split()[2]) # in Bars
            newline = outfile.readline() # C55
            moduli[4,4] = float(newline.strip().split()[2]) # in Bars
            newline = outfile.readline() # C66
            moduli[5,5] = float(newline.strip().split()[2]) # in Bars

            newline = outfile.readline() # C14
            moduli[0,3] = float(newline.strip().split()[2]) # in Bars
            moduli[3,0] = float(newline.strip().split()[2]) # in Bars
            newline = outfile.readline() # C15
            moduli[0,4] = float(newline.strip().split()[2]) # in Bars
            moduli[4,0] = float(newline.strip().split()[2]) # in Bars
            newline = outfile.readline() # C16
            moduli[0,5] = float(newline.strip().split()[2]) # in Bars
            moduli[5,0] = float(newline.strip().split()[2]) # in Bars

            newline = outfile.readline() # C24
            moduli[1,3] = float(newline.strip().split()[2]) # in Bars
            moduli[3,1] = float(newline.strip().split()[2]) # in Bars
            newline = outfile.readline() # C25
            moduli[1,4] = float(newline.strip().split()[2]) # in Bars
            moduli[4,1] = float(newline.strip().split()[2]) # in Bars
            newline = outfile.readline() # C26
            moduli[1,5] = float(newline.strip().split()[2]) # in Bars
            moduli[5,1] = float(newline.strip().split()[2]) # in Bars

            newline = outfile.readline() # C34
            moduli[2,3] = float(newline.strip().split()[2]) # in Bars
            moduli[3,2] = float(newline.strip().split()[2]) # in Bars
            newline = outfile.readline() # C35
            moduli[2,4] = float(newline.strip().split()[2]) # in Bars
            moduli[4,2] = float(newline.strip().split()[2]) # in Bars
            newline = outfile.readline() # C36
            moduli[2,5] = float(newline.strip().split()[2]) # in Bars
            moduli[5,2] = float(newline.strip().split()[2]) # in Bars

            newline = outfile.readline() # C45
            moduli[3,4] = float(newline.strip().split()[2]) # in Bars
            moduli[4,3] = float(newline.strip().split()[2]) # in Bars
            newline = outfile.readline() # C46
            moduli[3,5] = float(newline.strip().split()[2]) # in Bars
            moduli[5,3] = float(newline.strip().split()[2]) # in Bars
            newline = outfile.readline() # C56
            moduli[4,5] = float(newline.strip().split()[2]) # in Bars
            moduli[5,4] = float(newline.strip().split()[2]) # in Bars

            moduli_array.append(moduli)  # GPa
    outfile.close()
    return moduli_array



def _read_energy(outfilename):
    """
    Read single-point from a completed GULP job

    :param outfilename: Path of the stdout from the GULP job
    :type outfilename: str
    :returns: Energy of the structure in eV
    """
    energy_in_eV = []
    outfile = open(outfilename, 'r')
    for line in outfile:
        if 'EZFF_ENERGY' in line:
            energy_in_eV.append(float(line.strip().split()[-1]))
    outfile.close()
    return np.array(energy_in_eV)



def _read_structure(outfilename):
    """
    Read converged structure (cell and atomic positions) from the MD job

    :param outfilename: Path of file containing stdout of the GULP job
    :type outfilename: str

    :returns: xtal.AtTraj object with (optimized) individual structures as separate snapshots
    """
    relaxed = xtal.AtTraj()
    relaxed.box = np.zeros((3,3))

    # Read number of atoms, box definition and atom coordinates
    outfile = open(outfilename, 'r')
    for line in outfile:
        if 'NUMBER OF ATOMS' in line.strip():
            snapshot = relaxed.create_snapshot(xtal.Snapshot)
            nextline = outfile.readline()
            num_atoms = int(nextline.strip())
        elif 'BOX BOUNDS' in line.strip():
            if 'xy' in line.strip():
                l1 = outfile.readline()
                l2 = outfile.readline()
                l3 = outfile.readline()
                xlo_bound, xhi_bound, xy = list(map(float, l1.strip().split()))
                ylo_bound, yhi_bound, xz = list(map(float, l2.strip().split()))
                zlo_bound, zhi_bound, yz = list(map(float, l3.strip().split()))
                xlo = xlo_bound - np.amin([0.0, xy, xz, xy+xz])
                xhi = xhi_bound - np.amax([0.0, xy, xz, xy+xz])
                ylo = ylo_bound - np.amin([0.0, yz])
                yhi = yhi_bound - np.amax([0.0, yz])
                zlo = zlo_bound
                zhi = zhi_bound
                lx = xhi - xlo
                ly = yhi - ylo
                lz = zhi - zlo
                relaxed.abc = np.array([lx, np.sqrt((ly*ly)+(xy*xy)), np.sqrt((lz*lz)+(xz*xz)+(yz*yz))])
                cosa = ((xy*xz)+(ly*yz))/(relaxed.abc[1]*relaxed.abc[2])
                cosb = xz/relaxed.abc[2]
                cosc = xy/relaxed.abc[1]
                relaxed.ang = np.array([np.arccos(cosa), np.arccos(cosb), np.arccos(cosc)])
                relaxed.abc_to_box()
            else:
                l1 = outfile.readline()
                l2 = outfile.readline()
                l3 = outfile.readline()
                xlo, xhi = list(map(float, l1.strip().split()))
                ylo, yhi = list(map(float, l2.strip().split()))
                zlo, zhi = list(map(float, l3.strip().split()))
                lx = xhi - xlo
                ly = yhi - ylo
                lz = zhi - zlo
                relaxed.box[0][0] = lx
                relaxed.box[1][1] = ly
                relaxed.box[2][2] = lz
        elif 'ATOMS' in line.strip():
            for atomID in range(num_atoms):
                atom_details = outfile.readline().strip().split()
                atom = snapshot.create_atom(xtal.Atom)
                atom.element = atom_details[2].upper()
                atom.charge = float(atom_details[4])
                atom.cart = np.array(list(map(float,atom_details[5:8])))
                atom.vel = np.array(list(map(float,atom_details[8:11])))
                atom.force = np.array(list(map(float,atom_details[11:14])))

    outfile.close()
    return relaxed



"""This module provide functions for computing errors from previously completed MD runs"""
import os
import sys
import xtal
import numpy as np
# from platypus import Problem, unique, nondominated, NSGAII, NSGAIII, IBEA, PoolEvaluator
# from platypus.types import Real, Integer
# from platypus.operators import InjectedPopulation, GAOperator, SBX, PM



def error_phonon_dispersion(MD=None, GT=None, weights='uniform', verbose=False):
    """
    Calculate error between MD-computed phonon dispersion and the ground-truth phonon dispersion with user-defined weighting schemes

    :param MD: MD-computed phonon dispersion curve
    :type MD: 2D np.array

    :param GT: Ground-truth phonon dispersion curve
    :type GT: 2D np.array

    :param weights: User-defined weighting scheme for calculating errors provided as a list of numbers, one per band. Possible values are
                    ``uniform`` - where errors from all bands are equally weighted,
                    ``acoustic`` - where errors from lower-frequency bands are assigned greater weights, and
                    `list` - 1-D list of length equal to number of bands
    :type weights: str `or` list

    :param verbose: Deprecated option for verbosity of error calculation routine
    :type verbose: bool
    """
    # Perform sanity check. Number of bands should be equal between the two structures
    if not len(MD) == len(GT):
        raise ValueError("MD and ground truth dispersions have different number of bands")
        return np.nan

    # Create array of weights - one value per band
    num_band = len(MD)
    if weights == 'uniform':
        W = np.ones(num_band)
    elif weights == 'acoustic':
        maxfreq = np.amax(GT)
        W = np.reciprocal((np.mean(GT, axis=1)/maxfreq) + 0.1)
    elif isinstance(weights,list) or isinstance(weights,np.ndarray):
        if len(weights) == num_band:
            W = np.array(weights)
        else:
            raise ValueError("Number of provided weight values is different from number of bands! Aborting")

    # Compute the RMS error between dispersions
    try:
        rms_error = 0.0
        num_k_gt = len(GT[0])
        scaling = num_k_gt/100.0
        for band_index in range(0, len(GT)):
            interp_md_band = np.interp(np.arange(0, num_k_gt), np.arange(0, 100)*scaling, MD[band_index])
            rms_error += np.linalg.norm(interp_md_band - GT[band_index]) * W[band_index]

        rms_error /= (num_k_gt * num_band)
    except:
        rms_error = np.nan

    return rms_error



def error_structure_distortion(MD=None, GT=None):
    """
    Calculate error due to relaxation of atoms in the initial structure. The error is the sum of root mean square displacement of atoms.

    :param MD: Relaxed structure after MD run
    :type MD: xtal.AtTraj object

    :param GT: Initial Ground-Truth structure used as input for MD calculations
    :type GT: xtal.AtTraj object
    """
    # Sanity checks -- Both inputs should be AtTraj objects
    if not isinstance(MD, xtal.AtTraj) and isinstance(GT, xtal.AtTraj):
        print('ERROR_STRUCTURE_DISTORTION: Please provide xtal.AtTraj objects for comparison')
        return np.nan

    if not (len(GT.snaplist) == len(MD.snaplist)):
        print('Different number of structures in MD and Ground-Truth data')
        return np.nan

    if not np.allclose(MD.box, GT.box):
        print('ERROR: Cell sizes and shapes are different for MD and ground-truth. Please use error_lattice_constant instead')
        return np.nan

    try:
        error = 0.0
        for snapID in range(len(GT.snaplist)):
            errors_this_snapshot = []
            for atomID in range(len(GT.snaplist[snapID].atomlist)):
                error_this_atom = np.linalg.norm(GT.snaplist[snapID].atomlist[atomID].cart - MD.snaplist[snapID].atomlist[atomID].cart)
                errors_this_snapshot.append(error_this_atom)
            error += np.linalg.norm(np.array(errors_this_snapshot))
    except:
        error = np.nan

    return error



def error_lattice_constant(MD=None, GT=None):
    """
    Calculate error due to optimization of lattice constants in the initial structure.

    :param MD: Relaxed structure after MD run
    :type MD: xtal.AtTraj object

    :param GT: Initial Ground-Truth structure used as input for MD calculations
    :type GT: xtal.AtTraj object
    """
    # Sanity checks -- Both inputs should be AtTraj objects
    if not isinstance(MD, xtal.AtTraj) and isinstance(GT, xtal.AtTraj):
        print('ERROR_LATTICE_CONSTANT: Please provide xtal.AtTraj objects for comparison')
        return

    try:
        MD.make_dircar_matrices()
        GT.make_dircar_matrices()
        MD.box_to_abc()
        GT.box_to_abc()
        abc = MD.abc - GT.abc
        ang = MD.ang - GT.ang
    except:
        abc = np.array([np.nan, np.nan, np.nan])
        ang = np.array([np.nan, np.nan, np.nan])

    return abc, ang



def error_atomic_charges(MD=None, GT=None):
    """
    Calculate error due to difference between MD-computed atomic charges and ground-truth atomic charges

    :param MD: Relaxed structure after MD run
    :type MD: xtal.AtTraj object

    :param GT: Initial Ground-Truth structure used as input for MD calculations
    :type GT: xtal.AtTraj object
    """
    # Sanity checks -- Both inputs should be AtTraj objects
    if not isinstance(MD, xtal.AtTraj) and isinstance(GT, xtal.AtTraj):
        print('ERROR_ATOMIC_CHARGES: Please provide xtal.AtTraj objects for comparison')
        return np.nan

    if not (len(GT.snaplist) == len(MD.snaplist)):
        print('Different number of structures in MD and Ground-Truth data')
        return np.nan

    error_array = []
    for snapID in range(len(GT.snaplist)):
        GT_charges = np.array([atom.charge for atom in GT.snaplist[snapID].atomlist])
        MD_charges = np.array([atom.charge for atom in MD.snaplist[snapID].atomlist])
        try:
            error_this_snapshot = np.linalg.norm(GT_charges - MD_charges)
        except:
            return np.nan
        error_array.append(error_this_snapshot)

    return np.sum(error_array)



def error_energy(MD, GT, weights='uniform', verbose=False):
    """
    Calculate error between MD-computed potential energy surface and the ground-truth potential energy surface with user-defined weighting schemes

    :param md_disp: MD-computed potential energy surface
    :type md_disp: 1D np.array

    :param gt_disp: Ground-truth potential energy surface
    :type gt_disp: 1D np.array

    :param weights: User-defined weighting scheme for calculating errors. Possible values are
                    ``uniform`` - where errors from all points on the PES are weighted equally,
                    ``minima`` - where errors from lower-energy points are assigned greater weights,
                    ``dissociation`` - where errors from highest-energy points are assigned greater weights, and
                    `list` - 1-D list of length equal to number of points on the PES scans
    :type weights: str `or` list

    :param verbose: Deprecated option for verbosity of error calculation routine
    :type verbose: bool
    """
    # Perform sanity check. Number of bands should be equal between the two structures
    if not len(MD) == len(GT):
        raise ValueError("MD and ground truth PES have different number of points! Aborting")
        return

    MD = np.array(MD)
    GT = np.array(GT)

    num_pes = len(GT)
    W = np.ones(num_pes)
    if weights == 'uniform':
        pass
    elif weights == 'minima':
        min_E = np.amin(GT)
        max_E = np.amax(GT)
        W = np.reciprocal(((GT-min_E)/max_E) + 0.1)
    elif weights == 'dissociation':
        min_E = np.amin(GT)
        max_E = np.amax(GT)
        W = (9.0*(GT-min_E)/max_E) + 1.0
    elif isinstance(weights,list) or isinstance(weights,np.ndarray):
        if len(weights) == len(GT):
            W = np.array(weights)
        else:
            raise ValueError("Weights array and PES have different number of points! Aborting")

    # Compute the RMS error between PES
    try:
        rms_error = np.linalg.norm((MD - GT) * W)
    except:
        rms_error = np.nan

    return rms_error



import ezff
from ezff.interfaces import gulp, qchem
from ezff.utils.reaxff import reax_forcefield

# Define ground truths
gt_gs = qchem.read_structure('ground_truths/optCHOSx.out')
gt_gs_atomic_charges = qchem.read_atomic_charges('ground_truths/optCHOSx.out')

def my_error_function(variable_values, template):

    myrank = ezff.get_pool_rank()
    path = str(myrank)

    # Calculate Ground State Charges
    md_gs_job = gulp.job(path = path)
    md_gs_job.structure = gt_gs
    md_gs_job.forcefield = ezff.generate_forcefield(template, variable_values, FFtype = 'reaxff')
    md_gs_job.options['pbc'] = False
    md_gs_job.options['relax_atoms'] = False
    md_gs_job.options['relax_cell'] = False
    md_gs_job.options['atomic_charges'] = True
    # Run GULP calculation
    md_gs_job.run()
    # Read output from completed GULP job and clean-up
    md_gs_atomic_charges = md_gs_job.read_atomic_charges()
    md_gs_job.cleanup()

    # Calculate error
    charg_error = ezff.error_atomic_charges(MD=md_gs_atomic_charges, GT=gt_gs_atomic_charges)
    return [charg_error]


# Generate forcefield template and variable ranges
FF = reax_forcefield('ffield')
FF.make_template_qeq('S')
FF.generate_templates()


if __name__ == '__main__':

    obj = ezff.FFParam(error_function = my_error_function, num_errors = 1)
    obj.read_variable_bounds('param_ranges')
    obj.read_forcefield_template('ff.template.generated')

    obj.set_algorithm('randomsearch_so', population_size = 32)
    obj.parameterize(num_epochs = 5)
    obj.set_algorithm('ngopt_so', population_size = 32)
    obj.parameterize(num_epochs = 5)
